from importlib.resources import files
from time import sleep
from sys import stdout
from pint import UnitRegistry

from wxfrog import (
    start_gui, CalculationEngine, CalculationFailed, set_unit_registry)

my_registry = UnitRegistry()
set_unit_registry(my_registry)

my_registry.define('dog_year = 52 * day = dy')
my_registry.define('inverse_kilo_kelvin = 1e-3 / kelvin = tik = tik')
my_registry.define('barg = bar; offset: 1.01325 = barg = barg')
my_registry.define('kPag = kPa; offset: 101.325 = kPag = kPag')

Q = my_registry.Quantity

PAUSE_SECONDS = 0.1
NUM_ITER = 5

class MyModel(CalculationEngine):
    def __init__(self):
        self.outstream = stdout

    def initialise(self, out_stream):
        self.outstream = out_stream
        sleep(1)

    def get_default_parameters(self):
        return {"a": {"b": {"x": Q(3, "m**3/h")}},
                "c": Q(4, "bar")}

    def calculate(self, parameters: dict) -> dict:
        x = parameters["a"]["b"]["x"]
        if x < Q(1.5, "m^3/h"):
            raise CalculationFailed("Too little flow")
        c = Q(0.1, "bar*h/m**3")
        p1 = Q(30, "bar")
        p2 = p1 - c * x
        for i in range(NUM_ITER):
            sleep(PAUSE_SECONDS)
            print(f"Iteration {i} .. still doing nothing.", file=self.outstream)
        print(f"Now returning some fake values", file=self.outstream)

        more = {"Heater": {"Shell": {"U": Q(500, "W/(m^2*K)"),
                                     "Re": Q(43000),
                                     "Pr": Q(0.7)},
                           "Tube": {"U": Q(800, "W/(m^2*K)"),
                                    "Re": Q(73000),
                                    "Pr": Q(0.75)},
                            "duty": Q(25, "MW")},
                "Pump": {"power": Q(2, "MW"),
                         "efficiency": Q(87, "%")}
                }

        return {"y": Q(p2, "bar"),
                "z": Q(43.425, "degC"),
                "more": more,
                "streams": {
                    "s01": {
                        "T": Q(20, "degC"),
                        "p": p1,
                        "x": Q(0.7)},
                    "s02": {
                        "T": Q(40, "degC"),
                        "p": p2,
                        "x": Q(1e-5)}
                    }
                }


def main():
    model = MyModel()
    start_gui(files("wxfrog.examples.advanced.data"), model)

if __name__ == '__main__':
    main()




from importlib.resources import files
from wxfrog import start_gui, CalculationEngine, get_unit_registry

Q = get_unit_registry().Quantity


class MyModel(CalculationEngine):
    def get_default_parameters(self):
        return {"a": Q(1, "cm"), "b": Q(1, "cm")}

    def calculate(self, parameters: dict) -> dict:
        a, b = parameters["a"], parameters["b"]
        return {"A": a * b, "P": 2 * (a + b)}


def main():
    start_gui(
        files("wxfrog.examples.hello_world.data"),
        MyModel()
    )

if __name__ == '__main__':
    main()


from abc import ABC, abstractmethod
from io import TextIOBase

from wxfrog.utils import NestedQuantityMap, JSONType


class CalculationFailed(ValueError):
    """Exception to be thrown if the calculation fails"""
    pass

class CalculationEngine(ABC):
    def initialise(self, out_stream: TextIOBase):
        """This method is called initially to give the engine the opportunity
        to initialise itself.

        The initialisation is executed in a separate thread, so that the GUI
        is not blocked during the execution of it.

        :param out_stream: A thread-safe variant of a stream to receive output
          that will be displayed in the engine monitor.
        """
        pass

    def get_internal_state(self) -> JSONType:
        """The engine might have an internal state that is worth keeping track
        of for each scenario. The most obvious type of such would be a set of
        initial values for the independent variables of the underlying model.

        :return: A possibly nested data structure of json-serializable type,
          being sequences and mappings of strings, floats, ints, bools and None
          values
        """
        pass

    def set_internal_state(self, state: JSONType):
        """The counter-part of :meth:`get_internal_state`, receiving back the
        data structure to recover an internal state of the engine.

        :parameter state: The nested data structure representing the internal
          state
        """
        pass

    @abstractmethod
    def get_default_parameters(self) -> NestedQuantityMap:
        """Returns the default parameters of the underlying model as a nested
        map of pint quantities

        :return: The potentially nested map of default parameters
        """
        ...

    @abstractmethod
    def calculate(self, parameters: NestedQuantityMap) -> NestedQuantityMap:
        """
        Trigger a calculation of the underlying model, given the set of
        parameters and returning the results.

        This method is called in a separate thread to accommodate for the
        calculation to take significant time, during which the GUI then is not
        blocked.

        :param parameters: A potentially nested map of Quantity objects, whereas
          the structure is a subset of the structure returned by
          :meth:`get_default_parameters`. Unchanged parameters do not need to be
          provided by the client code.
        :return: A potentially nested map of (pint) quantities that represent
          the results of the calculation.
        """
        ...




from typing import Any, Callable
import wx
from wx.core import PyEventBinder

class PopupBase(wx.Dialog):
    """Abstract base class to open a transient window, covering a specific
    rectangle over the client control, and displaying any input control there.
    The operation is cancelled if the transient window loses focus. If a
    defined event is triggered by the input control, the data is extracted and
    sent back to parent context via a callback.
    """
    def __init__(self, parent: wx.Window, rect: wx.Rect,
                 initial_value, client_callback: Callable[[Any], bool]):
        super().__init__(parent, title="")
        self._client_callback = client_callback

        self._ctrl = self.create_ctrl(self, initial_value)
        min_size = self._ctrl.GetMinSize()
        size = rect.GetSize() + wx.Size(2, 2)
        min_size = wx.Size(max(min_size.GetWidth(), size.GetWidth()),
                           max(min_size.GetHeight(), size.GetHeight()))
        self._ctrl.SetMinSize(min_size)
        ds = min_size - size
        offset = wx.Point(4 + ds.GetWidth() // 2, 4 + ds.GetHeight() // 2)
        pos = rect.GetPosition() - offset
        self.SetPosition(pos)
        sizer = wx.BoxSizer(wx.VERTICAL)

        sizer.Add(self._ctrl, 1, wx.EXPAND | wx.ALL, 2)
        self.SetSizerAndFit(sizer)


    def bind_ctrl(self, event_type: PyEventBinder):
        """Bind an event to client control that triggers sending result
         back, and dismissing the popup"""
        self._ctrl.Bind(event_type, self._callback)

    def create_ctrl(self, parent: wx.Window, initial_value: Any) -> wx.Window:
        """Create and return control"""
        # cannot use ABC baseclass because of wx's SWIG meta-type of classes
        raise NotImplementedError("Abstract method call")

    def collect_result(self, event: wx.Event, ctrl: wx.Window) -> Any:
        """Collect the result from the control, as it is to be sent
        back to the parent code."""
        # cannot use ABC baseclass because of wx's SWIG meta-type of classes
        raise NotImplementedError("Abstract method call")

    def _dismiss(self, event=None):
        self.EndModal(wx.ID_ANY)
        wx.CallLater(200, self.Destroy)

    def _callback(self, event):
        result = self.collect_result(event, self._ctrl)
        if self._client_callback(result):
            self.EndModal(wx.ID_OK)


class APoint(wx.Point):
    @classmethod
    def from_point(cls, point: wx.Point):
        return cls(point.x, point.y)

    def __add__(self, size: wx.Size|wx.Point):
        return APoint(self.x + size.x, self.y + size.y)

    def __sub__(self, size: wx.Size|wx.Point):
        return APoint(self.x - size.x, self.y - size.y)

class ASize(wx.Size):
    def __floordiv__(self, other: int):
        return ASize(self.x // other, self.y // other)


from typing import Any
import wx
from .colors import ERROR_RED

class ValidatedTextCtrl(wx.Window):
    ERROR_SHOW_DURATION = 1500  # ms

    def __init__(self, parent, value: Any):
        super().__init__(parent)
        self.value = value

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.ctrl = wx.TextCtrl(self, value=self.format(value),
                                style=wx.TE_PROCESS_ENTER)
        sizer.Add(self.ctrl, 1, wx.EXPAND | wx.ALL, 2)
        self.SetSizerAndFit(sizer)
        self.SetMinSize(self.GetSize())

        self.ctrl.Bind(wx.EVT_TEXT_ENTER, self._on_value_changed)

    def _on_value_changed(self, event):
        ctrl = self.ctrl
        def reset():
            ctrl.SetBackgroundColour(orig_bg)
            ctrl.SetValue(self.format(self.value))

        orig_bg = self.ctrl.GetBackgroundColour()
        str_value = self.ctrl.GetValue()
        if not self.validate(str_value):
            ctrl.SetBackgroundColour(ERROR_RED)
            wx.CallLater(self.ERROR_SHOW_DURATION, reset)
        else:
            self.value = self.parse(str_value)
            event.Skip()

    def validate(self, value: str) -> bool:
        raise NotImplementedError("Pure virtual function call")

    def format(self, value: Any) -> str:
        raise NotImplementedError("Pure virtual function call")

    def parse(self, value: str) -> Any:
        raise NotImplementedError("Pure virtual function call")


class LogIncrementCtrl(ValidatedTextCtrl):
    def __init__(self, parent, value: float):
        self.parse = float
        self.format = lambda x: f"{value:.7g}"
        super().__init__(parent, value)

    def validate(self, value: str) -> bool:
        try:
            x = float(value)
        except ValueError:
            return False
        return x > 0 and x != 1.0

class NumberStepsCtrl(ValidatedTextCtrl):
    def __init__(self, parent, value: int):
        self.parse = int
        self.format = str
        super().__init__(parent, value)

    def validate(self, value: str) -> bool:
        try:
            n = int(value)
        except ValueError:
            return False
        return n > 0



from collections.abc import Mapping, Collection
from typing import Tuple
from datetime import datetime

from pubsub import pub
import wx
from wx.lib.mixins import listctrl
from wxfrog.models.scenarios import (
    SCENARIO_CURRENT, SCENARIO_CONVERGED, SCENARIO_DEFAULT)
from .colors import ERROR_RED, WARNING_ORANGE
from ..events import COPY_SCENARIO, RENAME_SCENARIO, DELETE_SCENARIO

class ScenarioNameDialog(wx.Dialog):
    def __init__(self, parent: wx.Window, names: Collection[str]):
        super().__init__(parent, title="Scenarios")
        self.names = names
        sizer = wx.BoxSizer(wx.VERTICAL)

        label=wx.StaticText(self, label="Enter name of scenario:")
        sizer.Add(label, 0, wx.EXPAND |wx.ALL, 5)

        self.name_ctrl = wx.TextCtrl(self)
        sizer.Add(self.name_ctrl, 0, wx.EXPAND | wx.ALL, 5)

        # status
        font = wx.Font(8, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL,
                       wx.FONTWEIGHT_NORMAL)
        self.status = wx.StaticText(self, label="")
        self.status.SetFont(font)
        sizer.Add(self.status, 0, wx.EXPAND | wx.LEFT, 3)

        sizer_2 = wx.BoxSizer(wx.HORIZONTAL)
        cancel = wx.Button(self, label="Cancel")
        self.ok = wx.Button(self, label="OK")
        self.ok.Enable(False)
        sizer_2.Add(cancel, 0, wx.ALL, 3)
        sizer_2.AddStretchSpacer(1)
        sizer_2.Add(self.ok, 0, wx.ALL, 3)
        sizer.Add(sizer_2, wx.EXPAND | wx.ALL, 3)
        self.SetSizer(sizer)
        self.Fit()

        self.name_ctrl.Bind(wx.EVT_TEXT, self._on_name_changed)
        self.ok.Bind(wx.EVT_BUTTON, lambda e: self.EndModal(wx.ID_OK))
        cancel.Bind(wx.EVT_BUTTON, lambda e: self.EndModal(wx.ID_CANCEL))

    @property
    def value(self):
        return self.name_ctrl.GetValue()

    def _on_name_changed(self, event):
        candidate = self.name_ctrl.GetValue()
        if candidate in self.names:
            self.status.SetForegroundColour(WARNING_ORANGE)
            self.status.SetLabel("Overwrite existing scenario?")
            self.ok.Enable(True)
        elif candidate.startswith("*"):
            self.status.SetForegroundColour(ERROR_RED)
            self.status.SetLabel("Own scenarios cannot start on '*'!")
            self.ok.Enable(False)
        elif not candidate:
            self.status.SetForegroundColour(ERROR_RED)
            self.status.SetLabel("You must pick a name!")
            self.ok.Enable(False)
        elif candidate.lower() in ("winter", "summer"):
            self.status.SetForegroundColour(WARNING_ORANGE)
            self.status.SetLabel("Seasonal greetings!")
            self.ok.Enable(True)
        else:
            self.status.SetLabel("")
            self.ok.Enable(True)


class ScenarioListCtrl(wx.ListCtrl, listctrl.ListCtrlAutoWidthMixin):
    def __init__(self, parent: wx.Window):
        super().__init__(parent, style=wx.LC_REPORT)
        listctrl.ListCtrlAutoWidthMixin.__init__(self)  # wx doesn't use super()
        self.SetMinSize(wx.Size(500, 200))
        self.InsertColumn(0, "Scenario", width=290)
        self.InsertColumn(1, "Modified", width=140)
        self.InsertColumn(2, "Results")


class ScenarioManager(wx.Dialog):
    def __init__(self, parent: wx.Window):
        super().__init__(parent, title="Manage Scenarios")
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.list = ScenarioListCtrl(self)
        sizer.Add(self.list, 1, wx.EXPAND | wx.ALL, 3)
        self.SetSizer(sizer)
        self.list.Bind(wx.EVT_LIST_ITEM_RIGHT_CLICK, self._on_right_click)
        self.SetSizerAndFit(sizer)

        self._popup_ids = None
        self._scenario_in_context = None

    def update(self, scenarios: Mapping[str, Tuple[bool, datetime]]):
        lst = self.list
        lst.DeleteAllItems()
        for name, (has_results, modified) in sorted(scenarios.items()):
            index = lst.InsertItem(lst.GetItemCount(), name)
            lst.SetItem(index, 1,  modified.strftime("%y-%m-%d %H:%M:%S"))
            lst.SetItem(index, 2,  "Yes" if has_results else "No")

    def _on_right_click(self, event):
        # define call-backs and condition when to show which items
        menu_items = {
            "Keep": (self._on_keep, (SCENARIO_CONVERGED, SCENARIO_CURRENT)),
            "Rename": (self._on_rename, ("custom",)),
            "Delete": (self._on_delete, ("custom",)),
            "Activate": (self._on_activate,
                         ("custom", SCENARIO_CONVERGED, SCENARIO_DEFAULT))}

        if self._popup_ids is None:
            ids = {}
            for n, (cb, _) in menu_items.items():
                ids[n] = wx.NewIdRef()
                self.Bind(wx.EVT_MENU, cb, id=ids[n])
            self._popup_ids = ids

        if not (name := event.GetText()):
            # when items changed, name can be empty, but event then fired twice
            return
        menu = wx.Menu()
        for n, ref_id in self._popup_ids.items():
            if (name if name[0] == "*" else "custom") in menu_items[n][1]:
                menu.Append(ref_id, n)

        self._scenario_in_context = name
        self.PopupMenu(menu)
        self._scenario_in_context = None
        menu.Destroy()

    def _on_keep(self, event):
        source = self._scenario_in_context
        names = self._get_custom_names()
        dialog = ScenarioNameDialog(self, names)
        if dialog.ShowModal() == wx.ID_OK:
            pub.sendMessage(COPY_SCENARIO, source=source, target=dialog.value)

    def _on_rename(self, event):
        source = self._scenario_in_context
        names = self._get_custom_names()
        dialog = ScenarioNameDialog(self, names)
        if dialog.ShowModal() == wx.ID_OK:
            pub.sendMessage(RENAME_SCENARIO, source=source, target=dialog.value)

    def _on_delete(self, event):
        source = self._scenario_in_context
        msg = f"Do you really want to delete the scenario\n'{source}'?"
        style = wx.YES_NO | wx.NO_DEFAULT
        dialog = wx.MessageDialog(self, msg, "Scenarios", style=style)
        if dialog.ShowModal() == wx.ID_YES:
            pub.sendMessage(DELETE_SCENARIO, name=source)

    def _on_activate(self, event):
        source = self._scenario_in_context
        pub.sendMessage(COPY_SCENARIO, source=source, target=SCENARIO_CURRENT)

    def _get_custom_names(self) -> Collection[str]:
        lst, idx, names = self.list, -1, set()
        while (idx := lst.GetNextItem(idx)) > -1:
            name = lst.GetItemText(idx)
            if not name.startswith("*"):
                names.add(name)
        return names


from typing import Self, Union, Optional
from collections.abc import Sequence, MutableMapping, Mapping
from io import TextIOBase, StringIO
from threading import Lock
from re import compile

import wx
from pint import UnitRegistry, DimensionalityError
from pint.registry import Quantity, Unit


JSONScalar = str | int | float | bool | None
JSONType = JSONScalar | Sequence["JSONType"] | Mapping[str, "JSONType"]
NestedStringMap = Mapping[str, Union[str, "NestedStringMap"]]
NestedQuantityMap = Mapping[str, Union[Quantity, "NestedQualityMap"]]
Path = tuple[str, ...]

_unit_registry: Optional[UnitRegistry] = None


def set_unit_registry(registry: UnitRegistry):
    global _unit_registry
    registry.autoconvert_offset_to_baseunit = True
    _unit_registry = registry


def get_unit_registry() -> UnitRegistry:
    if _unit_registry is None:
        set_unit_registry(UnitRegistry())
    return _unit_registry


def fmt_unit(unit: Unit):
    result = f"{unit:~P#}"
    return result.replace(" ", "")


class ThreadedStringIO(TextIOBase):
    def __init__(self):
        super().__init__()
        self._buf = StringIO()
        self._lock = Lock()
        self._buf_new = StringIO()

    def write(self, s):
        with self._lock:
            self._buf_new.write(s)
            return self._buf.write(s)

    def getvalue(self):
        with self._lock:
            return self._buf.getvalue()

    def get_recent(self):
        with self._lock:
            result = self._buf_new.getvalue()
            self._buf_new = StringIO()
            return result

    def flush(self):
        pass


class PathFilter:
    DOUBLE_STAR = r"([^.]+(\.[^.]+)<star>)"
    SINGLE_STAR = r"[^.]+"

    def __init__(self, search_term: str):
        """Create a filter with given search term. Examples are:

        - ``**.T``: Matches all paths ending with an element called ``T``.
          Here, ``**`` is a wildcard matching one or many arbitrary elements
          of the path
        - ``a.b.c.M``: Matches only the path as provided (no wildcards)
        - ``Synthesis.**.x.*``: Matches all paths that start with ``Synthesis``
          and have ``x`` as the second-last element, such as the path
          ``Synthesis/Reactor/Outlet/x/MeOH``.

        """
        if search_term:
            st = search_term.replace("\\*", "\\<star>")
            st = st.replace("**", self.DOUBLE_STAR)
            st = st.replace("*", self.SINGLE_STAR)
            st = st.replace("<star>", "*")
            self._pattern = compile(f"^{st}$")
            print(st)
        else:
            self._pattern = None

    def matches(self, path: Sequence[str]) -> bool:
        """Return ``True`` if the path - as is -  is matched by the search term.
        """
        if self._pattern is None:
            return True
        return self._pattern.match(".".join(path)) is not None


class DataStructure(dict, NestedQuantityMap):
    def get(self, path: Sequence[str]):
        res = self
        for p in path:
            res = res[p]
        return res

    def set(self, path: Sequence[str], value: Quantity):
        node = self.get(path[:-1])
        node[path[-1]] = value

    def to_jsonable(self) -> NestedStringMap:
        dive = self._dive(lambda x: f"{x:.14g~}")
        return dive(self)

    def convert_all_possible_to(self, unit: Unit):
        def dive(structure: MutableMapping):
           for k, v in structure.items():
               if isinstance(v, MutableMapping):
                   dive(v)
                   continue
               try:
                   structure[k] = v.to(unit)
               except DimensionalityError:
                   pass
        dive(self)

    @property
    def all_paths(self) -> Sequence[tuple[str, ...]]:
        def dive(struct: NestedQuantityMap, path: list[str]):
            if not isinstance(struct, Mapping):
                return True
            for k, v in struct.items():
                p = path + [k]
                if dive(v, p):
                    paths.append(tuple(p))
            return False

        paths = []
        dive(self, [])
        return paths

    @classmethod
    def from_jsonable(cls, nested_data: NestedStringMap) -> Self:
        dive = cls._dive(get_unit_registry().Quantity)
        return DataStructure(dive(nested_data))

    @staticmethod
    def _dive(func):
        def dive(struct):
            if isinstance(struct, Mapping):
                return {k: dive(v) for k, v in struct.items()}
            else:
                return func(struct)
        return dive

def copy_html_to_clipboard(html):
    def fill_clipboard():
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(data_obj)
            wx.TheClipboard.Close()
        else:
            wx.MessageDialog(None, "Clipboard not available", "Copy error",
                             style=wx.OK | wx.ICON_ERROR).ShowModal()

    # TODO: make it an option whether to copy in html or text format.
    #  Excel requires text, while LibreOffice requires html.
    if True:
        data_obj = wx.TextDataObject(html)
    else:
        html_format = wx.DataFormat(wx.DF_HTML)
        data_obj = wx.CustomDataObject(html_format)
        data_obj.SetData(html.encode("utf-8"))
    wx.CallLater(100, fill_clipboard)


import wx
from ..utils import ThreadedStringIO

class WatchTimer(wx.Timer):
    def __init__(self, owner, check_interval):
        super().__init__(owner)
        self.Start(milliseconds=check_interval)


class EngineMonitor(wx.Dialog):
    CHECK_INTERVAL = 500  # ms

    def __init__(self, parent: wx.Window, out_stream: ThreadedStringIO):
        style = wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER | wx.MAXIMIZE_BOX
        super().__init__(parent, title="Engine monitor", style=style)
        self.out_stream = out_stream

        sizer_1 = wx.BoxSizer(wx.VERTICAL)
        style = wx.TE_MULTILINE | wx.TE_READONLY | wx.HSCROLL
        self.text = wx.TextCtrl(self, style=style)
        font = wx.Font(wx.FontInfo(10).Family(wx.FONTFAMILY_TELETYPE))
        self.text.SetFont(font)
        self.text.SetMinSize(wx.Size(400, 300))
        sizer_1.Add(self.text, 1, wx.ALL | wx.EXPAND, 3)
        self.timer = WatchTimer(self, self.CHECK_INTERVAL)
        self.SetSizer(sizer_1)
        self.Fit()

        self.Bind(wx.EVT_TIMER, self._on_timer)

    def _on_timer(self, event):
        new_text = self.out_stream.get_recent()
        if new_text:
            new_value = self.text.GetValue() + new_text
            self.text.SetValue(new_value)
            self.text.ShowPosition(len(new_value) - 1)



from importlib.resources.abc import Traversable

import wx

from .controller import Controller
from .models.engine import CalculationEngine


class WxFrogApp(wx.App):
    def OnInit(self):
        return True

def start_gui(config_directory: Traversable, model: CalculationEngine):
    app = WxFrogApp()
    controller = Controller(config_directory, model)
    app.SetTopWindow(controller.frame)
    app.MainLoop()



import wx
from wx.html import HtmlWindow, EVT_HTML_LINK_CLICKED
from webbrowser import open

HTML = "<html><body>{content}</div></body></html>"


class AboutDialog(wx.Dialog):
    def __init__(self, parent: wx.Window, content: str, size: tuple[int, int]):
        super().__init__(parent, title="About",)
        sizer = wx.BoxSizer(wx.VERTICAL)
        html = HtmlWindow(self)
        content = HTML.format(content=content)
        html.SetPage(content)
        html.SetMinSize(wx.Size(*size))
        sizer.Add(html, 1, wx.EXPAND | wx.ALL, 3)
        sizer2 = wx.BoxSizer(wx.HORIZONTAL)
        footer = wx.StaticText(self, label="Powered by WxFrog, 2025")
        sizer2.Add(footer, 1, wx.ALIGN_CENTER_VERTICAL | wx.LEFT, 10)
        ok = wx.Button(self, label="Ok")
        sizer2.Add(ok, 0, wx.ALL, 3)
        ok.Bind(wx.EVT_BUTTON, lambda e: self.EndModal(wx.ID_OK))
        sizer.Add(sizer2, 0, wx.EXPAND, 0)
        self.SetSizerAndFit(sizer)

        html.Bind(EVT_HTML_LINK_CLICKED, self._link_clicked)

    @staticmethod
    def _link_clicked(event):
        url = event.GetLinkInfo().GetHref()
        open(url, new=2)



from dataclasses import dataclass, field
import wx

@dataclass
class TooltipInfo:
    item: dict = None
    counter: int = -1
    pos: wx.Point = field(default_factory=lambda: wx.Point(0, 0))
    pos_panel: wx.Point = field(default_factory=lambda: wx.Point(0, 0))

    def is_for(self, item):
        return self.item is not None and self.item["id"] == item["id"]

    def set_item(self, item):
        self.item = item
        self.counter = -1




from collections.abc import Mapping

import wx
from wx.dataview import (
    TreeListCtrl, TL_SINGLE, TL_NO_HEADER, TL_3STATE,
    EVT_TREELIST_ITEM_CHECKED)


class PropertyTreeListCtrl(TreeListCtrl):
    def __init__(self, parent: wx.Window):
        style = TL_SINGLE | TL_NO_HEADER | TL_3STATE
        super().__init__(parent, style=style)
        self.SetMinSize(wx.Size(400, 300))
        self.AppendColumn("check")
        self.AppendColumn("Path")

        a = self.AppendItem(self.GetRootItem(), "a")
        b = self.AppendItem(self.GetRootItem(), "b")
        c = self.AppendItem(self.GetRootItem(), "c")
        aa = self.AppendItem(a, "aa")
        ab = self.AppendItem(a, "ab")
        ac = self.AppendItem(a, "ac")
        ba = self.AppendItem(b, "ba")
        bb = self.AppendItem(b, "bb")
        bc = self.AppendItem(b, "bc")

        self.Bind(EVT_TREELIST_ITEM_CHECKED, self._on_item_checked)

    def set_paths(self, results):
        def add(item, structure):
            if not isinstance(structure, Mapping):
                return
            for k, v in structure.items():
                add(self.AppendItem(item, k), v)

        self.DeleteAllItems()
        add(self.GetRootItem(), results)

    @property
    def selected_paths(self):
        def collect(node, path):
            nonlocal paths
            state = self.GetCheckedState(node)
            if state == wx.CHK_CHECKED:
                paths.append(path)
            if state != wx.CHK_UNDETERMINED:
                return
            child = self.GetFirstChild(node)
            while child.IsOk():
                k = self.GetItemText(child)
                collect(child, path + [k])
                child = self.GetNextSibling(child)

        paths = []
        collect(self.GetRootItem(), [])
        return paths

    def _on_item_checked(self, event):
        #  - (de-)selecting checkbox updates status of parent check box
        #    depending on status of siblings
        #  - (de-)selecting checkbox updates all children to same status
        item = event.GetItem()
        state = self.GetCheckedState(item)
        self._update_parent_state(self.GetItemParent(item))
        self.CheckItemRecursively(item, state)

    def _update_parent_state(self, item):
        states = {wx.CHK_CHECKED: True, wx.CHK_UNCHECKED: False}
        selected = {True: False, False: False}
        child = self.GetFirstChild(item)
        while child.IsOk():
            try:
                s = states[self.GetCheckedState(child)]
            except KeyError:
                selected[True], selected[False] = True, True
            else:
                selected[s] = True
            child = self.GetNextSibling(child)
        if selected[True] and selected[False]:
            state = wx.CHK_UNDETERMINED
        else:
            state = wx.CHK_CHECKED if selected[True] else wx.CHK_UNCHECKED
        self.CheckItem(item, state)

        # update (grand-)parents
        if item != self.GetRootItem():
            parent = self.GetItemParent(item)
            self._update_parent_state(parent)



class PropertyPicker(wx.Dialog):
    def __init__(self, parent: wx.Window):
        style = wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER
        super().__init__(parent, title="Property selector", style=style)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.ctrl = PropertyTreeListCtrl(self)
        sizer.Add(self.ctrl, 1, wx.EXPAND | wx.ALL, 3)
        self.SetSizerAndFit(sizer)
        sizer2 = wx.BoxSizer(wx.HORIZONTAL)
        cancel = wx.Button(self, label="Cancel")
        sizer2.Add(cancel, 0, wx.EXPAND | wx.ALL, 3)
        sizer2.AddStretchSpacer(1)
        ok = wx.Button(self, label="Ok")
        sizer2.Add(ok, 0, wx.EXPAND | wx.ALL, 3)
        sizer.Add(sizer2, 0, wx.EXPAND | wx.ALL, 3)

        cancel.Bind(wx.EVT_BUTTON, lambda e: self.EndModal(wx.ID_CANCEL))
        ok.Bind(wx.EVT_BUTTON, lambda e: self.EndModal(wx.ID_OK))

    def set_paths(self, results):
        self.ctrl.set_paths(results)

    @property
    def selected_paths(self):
        return self.ctrl.selected_paths


from collections.abc import Collection
import wx
from wx.html import HtmlWindow

from ..config import ConfigurationError

_HEADER = """
    <b>Do not panic!</b><br><br> 
    The following errors are not hard to fix, but represent an inconsistency 
    between the connected model and the configuration. Maybe parameters or
    calculated properties have changed, and cannot be linked properly
    anymore. <br><br>
    Go through the individual errors and adjust the configuration file.
    """

class ConfigErrorDialog(wx.Dialog):
    def __init__(self, parent, errors: Collection[ConfigurationError]):
        style = wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER
        super().__init__(parent, title="Configuration errors",
                         style=style)
        sizer = wx.BoxSizer(wx.VERTICAL)
        header = HtmlWindow(self, style=wx.html.HW_SCROLLBAR_AUTO)
        header.SetPage(_HEADER)
        header.SetMinSize(wx.Size(600, 170))
        sizer.Add(header, 0, wx.EXPAND | wx.ALL, 3)

        style = (wx.TR_NO_LINES | wx.TR_HIDE_ROOT |
                 wx.TR_HAS_VARIABLE_ROW_HEIGHT | wx.TR_LINES_AT_ROOT)
        error_tree = wx.TreeCtrl(self, style=style)

        root = error_tree.AddRoot("Errors")
        for error in errors:
            item = error_tree.AppendItem(root, f"{error.message}:")
            error_tree.SetItemBold(item)
            error_tree.AppendItem(item, f"Path: {'.'.join(error.path)}")
            for key, value in error.details.items():
                error_tree.AppendItem(item, f"{key.capitalize()}: {value}")

        error_tree.ExpandAll()
        error_tree.SetMinSize(wx.Size(600, 150))
        sizer.Add(error_tree, 1, wx.EXPAND | wx.ALL, 3)

        sizer_2 = wx.BoxSizer(wx.HORIZONTAL)
        label = "The application will terminate when closing this dialog window"
        msg = wx.StaticText(self, label=label)
        sizer_2.Add(msg, 1, wx.ALIGN_CENTER_VERTICAL, 0)
        ok = wx.Button(self, label="OK")
        ok.Bind(wx.EVT_BUTTON, lambda e: self.EndModal(True))
        sizer_2.Add(ok, 0, wx.EXPAND |wx.ALL, 3)
        sizer.Add(sizer_2, 0, wx.EXPAND | wx.ALL, 3)

        self.SetSizer(sizer)
        self.Fit()




from abc import ABC, abstractmethod
import wx
from wx.svg import SVGimage


class ImageWrap(ABC):
    @property
    @abstractmethod
    def height(self):
        ...

    @property
    @abstractmethod
    def width(self):
        ...

    @abstractmethod
    def render_to_gc(self, gc, size):
        ...

class SVGImageWrap(ImageWrap):
    def __init__(self, file):
        image_data = bytes(file.read())
        self.image = SVGimage.CreateFromBytes(image_data)

    @property
    def height(self):
        return self.image.height

    @property
    def width(self):
        return self.image.width

    def render_to_gc(self, gc, size):
        self.image.RenderToGC(gc, size=size)

class PNGImageWrap(ImageWrap):
    def __init__(self, file):
        image = wx.Image(file, wx.BITMAP_TYPE_PNG)
        self.bitmap = wx.Bitmap(image)

    @property
    def height(self):
        return self.bitmap.GetHeight()

    @property
    def width(self):
        return self.bitmap.GetWidth()

    def render_to_gc(self, gc, size: wx.Size):
        w, h = size.GetWidth(), size.GetHeight()
        gc.DrawBitmap(self.bitmap, 0, 0, w, h)


from math import isfinite
from collections.abc import Sequence
from datetime import datetime
from html import escape
from locale import setlocale, localeconv, LC_ALL
from ..utils import get_unit_registry

setlocale(LC_ALL, "")

_SHELL = """
    <table style="border-collapse: collapse;">
      <tr><td nowrap>{label}:</td><td colspan="5">{title}</td></tr>
      <tr><td nowrap>Date:</td><td colspan="5">{date}</td></tr>
      <tr></tr>
    </table>
    <table style="border-collapse: collapse;">
      {columns}
      {table}
    </table>
"""
_CELL = "<td>{content}</td>"
_HEADING_CELL = "<td><b>{heading}</b></td>"
_BOTTOM_BORDER_STYLE = "border-bottom: thin solid black; "
_COL_BORDER_STYLE = """ style="border-right: thin solid black" """
_HEADING =  """<tr style="{border}background-color: {bg_col}">{row_hh}{col_hd}</tr>"""
_ROW = """<tr style="background-color: {bg_col}">{row_hh}{data}</tr>"""
_COL = "<col{style}/>"

_HTML_REPLACEMENTS = [
    ("°", "&deg;"),  ("µ", "&micro;"),  ("·", "&middot;"),  ("×", "&times;"),
    ("±", "&plusmn;"), ("Ω", "&Omega;"),  ("ω", "&omega;"),  ("Δ", "&Delta;"),
    ("δ", "&delta;"), ("σ", "&sigma;"), ("Å", "&Aring;")]

def recode(entity):
    result = escape(entity)
    return result.encode("ascii", "xmlcharrefreplace").decode("ascii")


class HtmlTable:
    HEADER_BG_COLOR = "#cccccc"
    ROW_BG_COLORS = [None, "#eeeeee"]

    def __init__(self, column_headers: Sequence[str],
                 row_headers: Sequence[str],
                 default_digits=6):
        self.label = "Simulation"
        self.title = ""
        self.date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self._column_headers = [column_headers]
        self._top_rect_headings = None
        self._row_headers = [row_headers]
        self._vertical_lines = set()
        self._row_format = {}
        self._col_format = {}
        self._default_digits = default_digits
        self._data = None
        self._threshold = [0.0] * len(row_headers)
        self._nan = [""] * len(row_headers)
        self._sub_threshold = [""] * len(row_headers)

    def add_column_header_row(self, row: Sequence[str]):
        assert len(row) == len(self._column_headers[0])
        self._column_headers.append(row)

    def add_row_header_column(self, column: Sequence[str]):
        assert len(column) == len(self._row_headers[0])
        self._row_headers.append(column)

    def add_vertical_line(self, column: int):
        self._vertical_lines.add(column)

    def add_row_unit_column(self, units: Sequence[str]):
        assert len(units) == len(self._row_headers[0])
        u_cls = get_unit_registry().Unit
        formatted = [f"{u_cls(u):H~}" for u in units]
        self.add_row_header_column(formatted)

    def add_column_unit_row(self, units: Sequence[str]):
        assert len(units) == len(self._column_headers[0])
        u_cls = get_unit_registry().Unit
        formatted = [f"{u_cls(u):H~}" for u in units]
        self.add_column_header_row(formatted)

    def set_top_rect_headers(self, headings: Sequence[Sequence[str]]):
        self._top_rect_headings = headings

    def set_row_format(self, row: int, fmt: str):
        self._row_format[row] = fmt

    def set_col_format(self, col: int, fmt: str):
        self._col_format[col] = fmt

    def set_threshold(self, row, threshold: float):
        assert threshold >= 0.0
        self._threshold[row] = threshold

    def set_nan(self, row, nan_string: str):
        self._nan[row] = nan_string

    def set_sub_threshold(self, row, sub_threshold_string: str):
        self._sub_threshold[row] = sub_threshold_string

    def set_data(self, data: Sequence[Sequence[float]]):
        def fmt(row: int, col: int, value: float):
            if not isfinite(value):
                return self._nan[row]
            if abs(value) < self._threshold[row]:
                return self._sub_threshold[row]

            f = self._row_format.get(row, "{x:.{d}g}")
            f = self._col_format.get(col, f)
            res = f.format(x=value, d=self._default_digits)
            return res.replace(".", dp)

        assert len(data) == len(self._row_headers[0])
        for d_r in data:
            assert len(d_r) == len(self._column_headers[0])
        dp = localeconv()["decimal_point"]
        self._data = [[fmt(r, c, d) for c, d in enumerate(d_r)]
                      for r, d_r in enumerate(data)]

    def render(self):
        assert self._data is not None
        headers = self._generate_headers()
        rows = self._generate_rows()
        columns = self._generate_columns()
        table = f"{headers}{rows}"
        result = _SHELL.format(
            label=self.label, title=self.title, date=self.date,
            columns=columns, table=table)

        # replace unicode entities from units of measurement
        for fr, to in _HTML_REPLACEMENTS:
            result = result.replace(fr, to)
        return result

    def _generate_columns(self):
        def style(k):
            return _COL_BORDER_STYLE if k in self._vertical_lines else ""

        cols = [_COL.format(style="")] * (len(self._row_headers) - 1)
        cols.append(_COL.format(style=_COL_BORDER_STYLE))
        cols.extend([
            _COL.format(style=style(k))
            for k in range(len(self._column_headers[0]))
        ])
        return "".join(cols)

    def _generate_rows(self):
        rows = []
        for k, row_h in enumerate(zip(*self._row_headers)):
            row_hh = "".join(_HEADING_CELL.format(heading=c) for c in row_h)
            data = "".join(_CELL.format(content=d) for d in self._data[k])
            bg_col = self.ROW_BG_COLORS[k % len(self.ROW_BG_COLORS)]
            rows.append(_ROW.format(bg_col=bg_col, row_hh=row_hh, data=data))
        return "\n".join(rows)

    def _generate_headers(self):
        rows = []
        last_k = len(self._column_headers) - 1
        top = ([[""] * len(self._row_headers)] * len(self._column_headers)
               if self._top_rect_headings is None else self._top_rect_headings)
        assert len(top) == len(self._column_headers)
        for k, r in enumerate(self._column_headers):
            left_hd = top[k]
            assert len(left_hd) == len(self._row_headers)
            row_hh = "".join(_HEADING_CELL.format(heading=c) for c in left_hd)
            headers = "".join(_HEADING_CELL.format(heading=c) for c in r)
            border = _BOTTOM_BORDER_STYLE if k == last_k else ""
            rows.append(_HEADING.format(
                border=border, bg_col=self.HEADER_BG_COLOR,
                row_hh=row_hh, col_hd=headers))
        return "\n".join(rows)





ACTIVATE_SCENARIO = "ACTIVATE_SCENARIO"
CALCULATION_DONE = "CALCULATION_DONE"
CALCULATION_FAILED = "CALCULATION_FAILED"
CASE_STUDY_ENDED = "CASE_STUDY_ENDED"
CASE_STUDY_INTERRUPT = "CASE_STUDY_INTERRUPT"
CASE_STUDY_LIST_CHANGED = "CASE_STUDY_LIST_CHANGED"
CASE_STUDY_NUMBER_CHANGED = "CASE_STUDY_NUMBER_CHANGED"
CASE_STUDY_PARAMETER_SELECTED = "CASE_STUDY_PARAMETER_SELECTED"
CASE_STUDY_PROGRESS = "CASE_STUDY_PROGRESS"
CASE_STUDY_PROPERTIES_SELECTED = "CASE_STUDY_PROPERTIES_SELECTED"
CASE_STUDY_RUN = "CASE_STUDY_RUN"
COPY_SCENARIO = "COPY_SCENARIO"
COPY_STREAM_TABLE = "COPY_STREAM_TABLE"
DELETE_SCENARIO = "DELETE_SCENARIO"
EXIT_APP = "EXIT_APP"
EXPORT_CANVAS_GFX = "EXPORT_CANVAS_GFX"
INITIALIZATION_DONE = "INITIALIZATION_DONE"
NEW_UNIT_DEFINED = "NEW_UNIT_DEFINED"
OPEN_FILE = "OPEN_FILE"
OPEN_RESULTS = "OPEN_RESULTS"
OPEN_SCENARIOS = "OPEN_SCENARIOS"
RENAME_SCENARIO = "RENAME_SCENARIO"
RESULT_UNIT_CHANGED = "RESULT_UNIT_CHANGED"
RESULT_UNIT_CLICKED = "RESULT_UNIT_CLICKED"
RUN_CASE_STUDY = "RUN_CASE_STUDY"
RUN_MODEL = "RUN_MODEL"
SAFE_FILE = "SAFE_FILE"
SAFE_FILE_AS = "SAFE_FILE_AS"
SHOW_PARAMETER_IN_CANVAS = "SHOW_PARAMETER_IN_CANVAS"



from typing import Callable
from collections.abc import Mapping, MutableMapping

import wx
from pubsub import pub
from pint.registry import Quantity

from wxfrog.models.casestudy import ParameterSpec
from wxfrog.events import (
    CASE_STUDY_PARAMETER_SELECTED, CASE_STUDY_LIST_CHANGED,
    CASE_STUDY_NUMBER_CHANGED, NEW_UNIT_DEFINED, CASE_STUDY_RUN,
    CASE_STUDY_PROGRESS, CASE_STUDY_INTERRUPT, CASE_STUDY_PROPERTIES_SELECTED,
    CASE_STUDY_ENDED)
from .auxiliary import PopupBase
from .quantity_control import (
    QuantityCtrl, QuantityChangedEvent, EVT_QUANTITY_CHANGED, EVT_UNIT_DEFINED)
from .number_ctrl import LogIncrementCtrl, NumberStepsCtrl
from .property_picker import PropertyPicker
from ..utils import DataStructure

class CaseProgressDialog(wx.ProgressDialog):
    _MSG = "Cases processed: {k}/{m}"

    def __init__(self, maximum: int):
        style = (wx.PD_CAN_ABORT | wx.PD_ELAPSED_TIME |
                 wx.PD_ESTIMATED_TIME | wx.PD_REMAINING_TIME | wx.PD_AUTO_HIDE)
        super().__init__(
            "Case study progress", self._MSG.format(k=0, m=maximum),
            maximum=maximum, parent=None, style=style)
        self._max = maximum
        pub.subscribe(self._update, CASE_STUDY_PROGRESS)
        pub.subscribe(self._destroy, CASE_STUDY_ENDED)

    def _destroy(self):
        pub.unsubscribe(self._update, CASE_STUDY_PROGRESS)
        pub.unsubscribe(self._destroy, CASE_STUDY_ENDED)
        self.Destroy()

    def _update(self, k: int):
        def do_update():
            res = False
            if k < self._max:
                res, _ = self.Update(k, self._MSG.format(k=k, m=self._max))
            if not res:
                pub.sendMessage(CASE_STUDY_INTERRUPT)
                self._destroy()
        wx.CallAfter(do_update)


class ParameterSelectDialog(wx.Dialog):
    def __init__(self, parent: wx.Window, parameters: DataStructure):
        super().__init__(parent, title="Select a parameter")
        sizer = wx.BoxSizer(wx.VERTICAL)
        style = wx.TR_HAS_BUTTONS | wx.TR_LINES_AT_ROOT | wx.TR_HIDE_ROOT
        self.tree = wx.TreeCtrl(self, style=style)
        self.tree.SetMinSize(wx.Size(500, 500))
        sizer.Add(self.tree, 1, wx.EXPAND | wx.ALL, 3)

        def populate(node, structure):
            for key, value in structure.items():
                child = self.tree.AppendItem(node, key)
                if isinstance(value, Mapping):
                    populate(child, value)

        root = self.tree.AddRoot("Root")
        populate(root, parameters)
        self.SetSizerAndFit(sizer)

        self.tree.Bind(wx.EVT_TREE_ITEM_ACTIVATED, self._on_activated)
        self.chosen = None

    def _on_activated(self, event):
        item = event.GetItem()
        if self.tree.ItemHasChildren(item):
            return  # can only select
        root = self.tree.GetRootItem()
        path = []
        while item != root:
            path.append(self.tree.GetItemText(item))
            item = self.tree.GetItemParent(item)
        self.chosen = tuple(reversed(path))
        self.EndModal(wx.ID_OK)


class NamePopup(PopupBase):
    def __init__(self, parent: wx.Window, value: str, callback, rect: wx.Rect):
        super().__init__(parent, rect, value, callback)
        self.bind_ctrl(wx.EVT_TEXT_ENTER)

    def create_ctrl(self, parent, initial_value):
        return wx.TextCtrl(parent, value=initial_value,
                           style=wx.TE_PROCESS_ENTER)

    def collect_result(self, event: wx.CommandEvent, ctrl):
        return event.GetString()


class QuantityPopup(PopupBase):
    def __init__(self, parent: wx.Window,
                 value: Quantity, callback, rect: wx.Rect, units,
                 min_value: Quantity = None, max_value: Quantity = None):
        self.min_value, self.max_value = min_value, max_value
        self.units = units
        super().__init__(parent, rect, value, callback)
        self.bind_ctrl(EVT_QUANTITY_CHANGED)

    def create_ctrl(self, parent, initial_value):
        def new_unit_defined(event):
            pub.sendMessage(NEW_UNIT_DEFINED, unit=event.new_unit)

        ctrl = QuantityCtrl(parent, initial_value, self.units,
                            min_value=self.min_value, max_value=self.max_value)
        ctrl.Bind(EVT_UNIT_DEFINED, new_unit_defined)
        return ctrl

    def collect_result(self, event: QuantityChangedEvent, ctrl):
        return event


class NumberPopup(PopupBase):
    def __init__(self, parent: wx.Window, value: int, callback, rect: wx.Rect):
        super().__init__(parent, rect, value, callback)
        self.bind_ctrl(wx.EVT_TEXT_ENTER)

    def create_ctrl(self, parent, initial_value):
        return NumberStepsCtrl(parent, initial_value)

    def collect_result(self, event: QuantityChangedEvent,
                       ctrl: NumberStepsCtrl):
        return int(ctrl.value)


class FactorPopup(PopupBase):
    def __init__(self, parent: wx.Window, value: float,
                 callback, rect: wx.Rect):
        super().__init__(parent, rect, value, callback)
        self.bind_ctrl(wx.EVT_TEXT_ENTER)

    def create_ctrl(self, parent, initial_value):
        return LogIncrementCtrl(parent, initial_value)

    def collect_result(self, event: QuantityChangedEvent, ctrl):
        return float(event.GetString())


class ParameterListCtrl(wx.ListCtrl):
    def __init__(self, parent: wx.Window):
        style = wx.LC_REPORT | wx.LC_SINGLE_SEL
        super().__init__(parent, style=style)

        self.columns = [["Path", 150], ["Name", 150], ["Min", 100],
                        ["Max", 100], ["Increment", 100], ["Steps", 100],
                        ["Log", 40]]
        for k, (n, w) in enumerate(self.columns):
            self.InsertColumn(k, n, width=w)
        self.parameters : list[MutableMapping] = []

        width = sum(w for _, w in self.columns)
        self.SetMinSize(wx.Size(width, 300))
        self.Bind(wx.EVT_SIZE, self._on_size)
        self.Bind(wx.EVT_LIST_COL_END_DRAG, self._on_column_resized)
        self.Bind(wx.EVT_LEFT_DCLICK, self._on_item_activated)

        cb = lambda e: pub.sendMessage(CASE_STUDY_LIST_CHANGED)
        for evt in (wx.EVT_LIST_ITEM_SELECTED, wx.EVT_LIST_DELETE_ITEM,
                    wx.EVT_LIST_INSERT_ITEM, wx.EVT_LIST_ITEM_DESELECTED):
            self.Bind(evt, cb)

    def set_parameters(self, param: list[MutableMapping]):
        self.parameters = []
        self.DeleteAllItems()
        for p in param:
            self.add(p)

    def get_parameters(self) -> list[MutableMapping]:
        return self.parameters

    def _on_item_activated(self, event):
        pos = event.GetPosition()
        item, flags, subitem = self.HitTestSubItem(pos)
        if item == wx.NOT_FOUND or subitem == wx.NOT_FOUND:
            return
        rect = wx.Rect()
        self.GetSubItemRect(item, subitem, rect, wx.LIST_RECT_BOUNDS)

        callbacks : Mapping[int, Callable[[int, wx.Rect], None]] = {
            1: self._on_edit_name,
            2: self._on_edit_min,
            3: self._on_edit_max,
            4: self._on_edit_incr,
            5: self._on_edit_number,
            6: self._on_toggle_log
        }
        if subitem in callbacks:
            callbacks[subitem](item, rect)

    def _on_edit_name(self, item, rect):
        def commit(value):
            self.SetItem(item, 1, value)
            return True

        name = self.GetItemText(item, col=1)
        popup = NamePopup(self, name, commit, rect)
        wx.CallAfter(popup.ShowModal)

    def _on_edit_min(self, item, rect):
        def commit(event):
            new_value = event.new_value
            if not event.in_bounds or new_value == spec.min:
                return False
            incr, num = None, None
            if spec.num_spec:
                num = spec.num
            else:
                incr = spec.incr
            try:
                info["spec"] = ParameterSpec(
                    spec.path, event.new_value, spec.max, name=spec.name,
                    num=num, incr=incr, log=spec.log)
            except TypeError:
                wx.MessageBox("Log mode not feasible for sign-shifting interval",
                              "Error in case study",
                              style=wx.ICON_ERROR | wx.CANCEL)
                return False

            self.update(item)
            return event.enter_pressed()

        info = self.parameters[item]
        spec = info["spec"]
        popup = QuantityPopup(self, spec.min, commit, rect,
                              info["units"], info["min"], info["max"])
        wx.CallAfter(popup.ShowModal)

    def _on_edit_max(self, item, rect):
        def commit(event):
            new_value = event.new_value
            if not event.in_bounds or new_value == spec.max:
                return False
            incr, num = None, None
            if spec.num_spec:
                num = spec.num
            else:
                incr = spec.incr
            try:
                info["spec"] = ParameterSpec(
                    spec.path, spec.min, event.new_value, name=spec.name,
                    num=num, incr=incr, log=spec.log)
            except TypeError:
                wx.MessageBox("Log mode not feasible for sign-shifting interval",
                              "Error in case study",
                              style=wx.ICON_ERROR | wx.CANCEL)
                return False

            self.update(item)
            return event.enter_pressed()

        info = self.parameters[item]
        spec = info["spec"]
        popup = QuantityPopup(self, spec.max, commit, rect,
                              info["units"], info["min"], info["max"])
        wx.CallAfter(popup.ShowModal)

    def _on_edit_incr(self, item, rect):
        def commit_incr(event):
            incr = event.new_value
            info["spec"] = ParameterSpec(
                spec.path, spec.min, spec.max, name=spec.name,
                incr=incr, log=spec.log)
            self.update(item)
            return event.enter_pressed()

        def commit_factor(factor):
            info["spec"] = ParameterSpec(
                spec.path, spec.min, spec.max, name=spec.name,
                incr=factor, log=spec.log)
            self.update(item)
            return True

        info = self.parameters[item]
        spec = info["spec"]
        if spec.log:
            # number popup (need to make, number needs to be positive)
            popup = FactorPopup(self, spec.incr, commit_factor, rect)
        else:
            popup = QuantityPopup(self, spec.incr, commit_incr,
                                  rect, info["units"])

        wx.CallAfter(popup.ShowModal)

    def _on_edit_number(self, item, rect):
        def commit(number: int):
            info["spec"] = ParameterSpec(
                spec.path, spec.min, spec.max, name = spec.name,
                num=number, log=spec.log)
            self.update(item)
            return True

        info = self.parameters[item]
        spec = info["spec"]
        popup = NumberPopup(self, spec.num, commit, rect)
        wx.CallAfter(popup.ShowModal)

    def _on_toggle_log(self, item, rect):
        info = self.parameters[item]
        spec = info["spec"]
        try:
            info["spec"] = ParameterSpec(
                spec.path, spec.min, spec.max, name=spec.name,
                num=spec.num, log=not spec.log)
        except TypeError:
            wx.MessageBox("Log mode not feasible for sign-shifting interval",
                          "Error in case study",
                          style=wx.ICON_ERROR | wx.CANCEL)
        else:
            self.update(item)

    def _on_size(self, event):
        size = event.GetSize()
        width = sum(w for _, w in self.columns)
        self.columns = [[n, w / width * size[0]] for n, w in self.columns]
        for k, (_, w) in enumerate(self.columns):
            self.SetColumnWidth(k, int(w))
        event.Skip()

    def _on_column_resized(self, event):
        new_w = [self.GetColumnWidth(i) for i in range(len(self.columns))]
        dw = sum(nw - w for nw, (_, w) in zip(new_w, self.columns))
        new_w[-1] = max(int(new_w[-1] - dw), 10)
        self.SetColumnWidth(len(self.columns) - 1, new_w[-1])
        self.columns = [[n, nw] for (n, _), nw in zip(self.columns, new_w)]

    def add(self, param_info: MutableMapping):
        idx = self.GetItemCount()
        spec = param_info["spec"]
        self.InsertItem(idx, ".".join(spec.path))
        self.parameters.append(param_info)
        self.update(idx)

    def update(self, idx):
        spec = self.parameters[idx]["spec"]
        self.SetItem(idx, 0, ".".join(spec.path))
        self.SetItem(idx, 1, spec.name)
        self.SetItem(idx, 2, f"{spec.min:.6g~P}")
        self.SetItem(idx, 3, f"{spec.max:.6g~P}")
        incr = f"{spec.incr:.6g}" if spec.log else f"{spec.incr:.6g~P}"
        self.SetItem(idx, 4, incr)
        self.SetItem(idx, 5, f"{spec.num}")
        self.SetItem(idx, 6, "Yes" if spec.log else "No")

        pub.sendMessage(CASE_STUDY_NUMBER_CHANGED, number=self.total_number)

    @property
    def total_number(self):
        if not self.parameters:
            return -1
        total_number = 1
        for info in self.parameters:
            total_number *= info["spec"].num
        return total_number


class CaseStudyDialog(wx.Dialog):
    _TOTAL_NUMBER_MSG = "Total number of cases to run: "
    def __init__(self, parent: wx.Window):
        style = wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER
        super().__init__(parent, title="Case study", style=style)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.list_ctrl = ParameterListCtrl(self)
        sizer.Add(self.list_ctrl, 1, wx.EXPAND | wx.ALL, 3)

        icon_size = wx.Size(16, 16)
        sizer_2 = wx.BoxSizer(wx.HORIZONTAL)

        btn_data = [("add", wx.ART_PLUS, False, self._on_add),
                    ("up", wx.ART_GO_UP, False, self._on_up),
                    ("down", wx.ART_GO_DOWN, False, self._on_down),
                    ("del", wx.ART_CROSS_MARK, False, self._on_delete),
                    ("copy", wx.ART_COPY, False, self._on_copy_results)]
        self.buttons = {}
        for name, icon, enabled, call_back in btn_data:
            bmp = wx.ArtProvider.GetBitmap(icon , wx.ART_BUTTON, icon_size)
            btn = wx.BitmapButton(self, bitmap=wx.BitmapBundle(bmp))
            btn.Enable(enabled)
            btn.Bind(wx.EVT_BUTTON, call_back)
            self.buttons[name] = btn

        (run_btn := wx.Button(self, label="Run")).Enable(False)
        run_btn.Bind(wx.EVT_BUTTON, self._on_run)
        self.buttons["run"] = run_btn

        for name in "add up down del".split():
            sizer_2.Add(self.buttons[name], 0, wx.EXPAND | wx.ALL, 3)
        label = f"{self._TOTAL_NUMBER_MSG } -"
        self.total_number_label = wx.StaticText(self, label=label)
        pub.subscribe(self._on_total_number_changed, CASE_STUDY_NUMBER_CHANGED)
        sizer_2.Add(self.total_number_label, 1,
                    wx.ALL | wx.ALIGN_CENTER_VERTICAL, 3)
        for name in "copy run".split():
            sizer_2.Add(self.buttons[name], 0, wx.EXPAND | wx.ALL, 3)
        sizer.Add(sizer_2, 0, wx.EXPAND, 0)
        self.SetSizerAndFit(sizer)
        pub.subscribe(self._on_list_changed, CASE_STUDY_LIST_CHANGED)

        self._property_picker = PropertyPicker(self)
        self._scenario = None
        self._allow_run = False
        self._progress = None


    def switch_button_enable(self, name: str, enabled: bool):
        self.buttons[name].Enable(enabled)

    def set_scenario(self, scenario):
        self._scenario = scenario

    def _select(self, item):
        self.list_ctrl.SetItemState(
            item - 1, wx.LIST_STATE_SELECTED, wx.LIST_STATE_SELECTED)

    def _on_run(self, event):
        specs = [p["spec"] for p in self.list_ctrl.parameters]
        pub.sendMessage(CASE_STUDY_RUN, specs=specs)
        self._progress = CaseProgressDialog(self.list_ctrl.total_number)

    def _on_total_number_changed(self, number):
        num_fmt = "-" if number < 0 else str(number)
        msg = f"{self._TOTAL_NUMBER_MSG}{num_fmt}"
        self.total_number_label.SetLabel(msg)

    def _on_add(self, event):
        # show tree dialog with parameters to select from
        param = self._scenario.parameters
        dialog = ParameterSelectDialog(self, param)
        if dialog.ShowModal() != wx.ID_OK:
            return
        path = dialog.chosen
        for info in self.list_ctrl.parameters:
            spec = info["spec"]
            if spec.path == path:
                id_ = ".".join(path)
                wx.MessageDialog(
                    self, f"Parameter '{spec.name}' ({id_}) already selected",
                    "Parameter selection error", style=wx.ICON_ERROR | wx.OK
                ).ShowModal()
                return
        pub.sendMessage(CASE_STUDY_PARAMETER_SELECTED, path=path)

    def _on_delete(self, event):
        lc = self.list_ctrl
        item = lc.GetFirstSelected()
        lc.DeleteItem(item)
        del lc.parameters[item]

    def _on_up(self, event):
        lc = self.list_ctrl
        item = lc.GetFirstSelected()
        lc.parameters[item], lc.parameters[item - 1] = \
            lc.parameters[item - 1], lc.parameters[item]
        lc.update(item)
        lc.update(item - 1)
        self._select(item - 1)

    def _on_down(self, event):
        lc = self.list_ctrl
        item = lc.GetFirstSelected()
        lc.parameters[item], lc.parameters[item + 1] = \
            lc.parameters[item + 1], lc.parameters[item]
        lc.update(item)
        lc.update(item + 1)
        self._select(item + 1)

    def _on_list_changed(self):
        item = self.list_ctrl.GetFirstSelected()
        count = self.list_ctrl.GetItemCount()
        if item == -1:
            for name in "up down del".split():
                self.buttons[name].Enable(False)
        else:
            self.buttons["del"].Enable(True)
            self.buttons["up"].Enable(item > 0)
            self.buttons["down"].Enable(item < count - 1)
        self._update_run_button_status()

    def _on_copy_results(self, event):
        # select properties based on current selection
        picker = self._property_picker
        picker.set_paths(self._scenario.results)
        if picker.ShowModal() == wx.ID_OK:
            pub.sendMessage(CASE_STUDY_PROPERTIES_SELECTED,
                            paths=picker.selected_paths)

    def _update_run_button_status(self):
        param_defined = (self.list_ctrl.GetItemCount() > 0)
        self.switch_button_enable("run", self._allow_run and param_defined)

    def allow_run(self, enable: bool):
        self._allow_run = enable
        self._update_run_button_status()
        self.switch_button_enable("copy", enable)



from datetime import UTC, datetime
from collections.abc import Sequence
from typing import Self
from pint.registry import Quantity

from ..utils import DataStructure, JSONType


SCENARIO_DEFAULT = "* Default"
SCENARIO_CURRENT = "* Active"
SCENARIO_CONVERGED = "* Converged"


class Scenario:
    def __init__(self, parameters: DataStructure, modified: datetime = None):
        self.parameters: DataStructure = parameters
        self.results: DataStructure = DataStructure()
        self.internal_state : JSONType = None
        self.modified = datetime.now(UTC) if modified is None else modified

    def set_param(self, path: Sequence[str], value: Quantity):
        self.parameters.set(path, value)
        self.modified = datetime.now(UTC)
        self.results = DataStructure()

    def mod_local_time(self):
        return self.modified.astimezone()

    def has_results(self) -> bool:
        return len(self.results) > 0

    def serialize(self) -> JSONType:
        return {
            "parameters": self.parameters.to_jsonable(),
            "state": self.internal_state,
            "results": self.results.to_jsonable(),
            "modified": self.modified.isoformat()
        }

    @classmethod
    def deserialize(cls, data: JSONType) -> Self:
        result = cls(DataStructure.from_jsonable(data["parameters"]),
                     datetime.fromisoformat(data["modified"]))
        result.internal_state = data["state"]
        result.results = DataStructure.from_jsonable(data["results"])
        return result



"""Simple module to define the version - such that it can be read from other
places as well."""

MAJOR = 1  #: Major version
MINOR = 0  #: Minor version
FLAG = "beta"  #: Version flag (None if release, else alpha or beta)
BUILD = 1  #: Build flag

version = f"{MAJOR}.{MINOR}"
if FLAG is not None:
    version = f"{version}-{FLAG[0]}{BUILD}"



from collections.abc import Collection
from pubsub.pub import sendMessage
import wx

from ..config import Configuration, ConfigurationError
from .canvas import Canvas
from .config_error_dialog import ConfigErrorDialog
from .engine_monitor import EngineMonitor
from .scenario import ScenarioManager
from .results import ResultView
from .casestudy import CaseStudyDialog
from .about import AboutDialog
from ..events import (
    EXPORT_CANVAS_GFX, RUN_MODEL, OPEN_SCENARIOS, OPEN_FILE, SAFE_FILE,
    SAFE_FILE_AS, EXIT_APP, RUN_CASE_STUDY, OPEN_RESULTS, COPY_STREAM_TABLE)
from ..utils import ThreadedStringIO

_FD_STYLE_LOAD = wx.FD_OPEN | wx.FD_FILE_MUST_EXIST
_FD_STYLE_SAVE = wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT | wx.FD_CHANGE_DIR

wx.Log.SetLogLevel(0)  # prevent unnecessary warnings


class FrogFrame(wx.Frame):
    def __init__(self, config: Configuration, out_stream: ThreadedStringIO):
        self.config = config
        title = f"{config['app_name']}"
        super().__init__(None, title=title)
        self._icon = config.get_app_icon()
        if self._icon is not None:
            self.SetIcon(self._icon)
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.canvas = Canvas(self, self.config)
        sizer.Add(self.canvas, 1, wx.EXPAND | wx.ALL, 3)
        self.run_menu_item = None
        self.case_study_menu_item = None
        self.copy_stream_table_menu_item = None
        self.define_menu()

        self.monitor = EngineMonitor(self, out_stream)
        self.scenarios = ScenarioManager(self)
        self.results = ResultView(self)
        self.case_studies = CaseStudyDialog(self)
        about_text = config.get("about", f"<h3>{config['app_name']}</h3>")
        about_size = config.get("about_size", (400, 200))
        self._about = AboutDialog(self, about_text, about_size)

        self.SetSizerAndFit(sizer)

        desired_size = wx.Size(self.canvas.bg_size)  # make a copy
        display = wx.Display().GetGeometry()
        delta = 100
        if desired_size.GetHeight() > display.GetHeight() - delta:
            desired_size.SetHeight(display.GetHeight() - delta)
        if desired_size.GetWidth() > display.GetWidth() - delta:
            desired_size.SetWidth(display.GetWidth() - delta)

        def adjust_size():
            for _ in range(3):
                ds = desired_size - self.canvas.GetClientSize()
                self.SetSize(self.GetSize() + ds)
            self.SetMaxSize(self.GetSize())

        wx.CallLater(100, adjust_size)

    def define_menu(self):
        menu_bar = wx.MenuBar()

        file_menu = wx.Menu()
        item = file_menu.Append(wx.ID_ANY, "&Open\tCTRL+O", "Open file")
        self.Bind(wx.EVT_MENU, lambda e: sendMessage(OPEN_FILE), item)
        item = file_menu.Append(wx.ID_ANY, "&Save\tCTRL+S", "Save file")
        self.Bind(wx.EVT_MENU, lambda e: sendMessage(SAFE_FILE), item)
        item = file_menu.Append(wx.ID_ANY, "Save &As ...\tCTRL+W",
                                "Save file with another name")
        self.Bind(wx.EVT_MENU, lambda e: sendMessage(SAFE_FILE_AS), item)
        item = file_menu.Append(wx.ID_ANY, "&Export canvas ...\tCTRL+G",
                                "Export canvas as png")
        self.Bind(wx.EVT_MENU, lambda e: sendMessage(EXPORT_CANVAS_GFX), item)
        item = file_menu.Append(wx.ID_ANY, "&Copy stream table\tCTRL+T",
                                "Copy stream table to clipboard (for Excel)")
        self.Bind(wx.EVT_MENU, self._on_copy_stream_table, item)
        item.Enable(False)
        self.copy_stream_table_menu_item = item
        item = file_menu.Append(wx.ID_ANY, "E&xit\tCTRL+X", "Exit simulator")
        self.Bind(wx.EVT_MENU, lambda e: self.Close(), item)
        menu_bar.Append(file_menu, "&File")

        engine_menu = wx.Menu()
        item = engine_menu.Append(wx.ID_ANY, "&Run model\tCTRL+R",
                                  "Run model in background")
        self.Bind(wx.EVT_MENU, lambda e: sendMessage(RUN_MODEL), item)
        item.Enable(False)
        self.run_menu_item = item
        item = engine_menu.Append(wx.ID_ANY, "&Case study ...\tCTRL+C",
                                 "Run case study")
        item.Enable(False)
        self.case_study_menu_item = item
        self.Bind(wx.EVT_MENU, lambda e: sendMessage(RUN_CASE_STUDY), item)
        menu_bar.Append(engine_menu, "&Engine")

        view_menu = wx.Menu()
        item = view_menu.Append(wx.ID_ANY, "&Monitor\tCTRL+E",
                                  "Show engine monitor")
        self.Bind(wx.EVT_MENU, lambda e: self.monitor.Show(), item)
        item = view_menu.Append(wx.ID_ANY, "&Scenarios\tCTRL+M",
                                 "Manage Scenarios")
        self.Bind(wx.EVT_MENU, lambda e: sendMessage(OPEN_SCENARIOS), item)
        item = view_menu.Append(wx.ID_ANY, "&All Results\tCTRL+A",
                                 "Show results")
        self.Bind(wx.EVT_MENU, lambda e: sendMessage(OPEN_RESULTS), item)
        menu_bar.Append(view_menu, "&View")

        help_menu = wx.Menu()
        item = help_menu.Append(wx.ID_ANY, "&About", "About this application")
        self.Bind(wx.EVT_MENU, lambda e: self._about.ShowModal(), item)
        menu_bar.Append(help_menu, "&Help")

        self.SetMenuBar(menu_bar)
        self.Bind(wx.EVT_CLOSE, lambda e: sendMessage(EXIT_APP, event=e))

    def show_calculation_error(self, error: str):
        self.canvas.set_results_mode(Canvas.RESULT_ERROR)
        self.Refresh()
        msg = f"{error}\n\nDo you want to open the engine monitor?"
        style = wx.YES_NO | wx.ICON_ERROR
        title = "Engine calculation error"
        dialog = wx.MessageDialog(self, msg, title, style=style)
        if dialog.ShowModal() == wx.ID_YES:
            self.monitor.Show()

    def show_file_dialog(self, msg: str, file_type: str, ending: str,
                         save: bool):

        wildcard = f"{file_type} files (*.{ending})|*.{ending}"
        style = _FD_STYLE_SAVE if save else _FD_STYLE_LOAD
        dialog = wx.FileDialog(self, msg, wildcard=wildcard, style=style)
        if dialog.ShowModal() == wx.ID_OK:
            path = dialog.GetPath()
            if path.lower().endswith(f".{ending}"):
                return path
            else:
                return f"{path}.{ending}"
        else:
            return None

    def update_title(self, filename: str):
        title = f"{self.config['app_name']} - {filename}"
        self.SetTitle(title)

    def show_config_error_dialog(self, errors: Collection[ConfigurationError]):
        dialog = ConfigErrorDialog(self, errors)
        dialog.ShowModal()
        self.Close()

    def show_case_studies(self, scenario):
        cs = self.case_studies
        cs.set_scenario(scenario)
        cs.Show()

    def _on_copy_stream_table(self, event):
        table_def = self.config.get("tables", {})
        match len(table_def):
            case 0:
                wx.MessageBox("No stream table defined", "Export stream table",
                              wx.ICON_WARNING | wx.CANCEL)
            case 1:
                first = next(iter(table_def))
                sendMessage(COPY_STREAM_TABLE, name=first)
            case _:
                # TODO: simple list dialog (Combobox) to select a table to copy.
                print("Multiple stream tables are not yet supported")



from pubsub import pub
from pint.registry import Quantity
import wx

from ..events import NEW_UNIT_DEFINED

from .quantity_control import (
    QuantityCtrl, EVT_QUANTITY_CHANGED, EVT_UNIT_DEFINED)


class ParameterDialog(wx.Dialog):
    def __init__(self, parent: wx.Window, item, value, units):
        super().__init__(parent, title=item['name'])
        sizer_1 = wx.BoxSizer(wx.VERTICAL)
        q_ctrl = QuantityCtrl(self, value, units,
                              item.get("min", None), item.get("max", None))
        sizer_1.Add(q_ctrl, 0, wx.EXPAND | wx.TOP | wx.RIGHT | wx.LEFT, 3)

        sizer_2 = wx.BoxSizer(wx.HORIZONTAL)
        cancel = wx.Button(self, label="Cancel")
        self.ok = wx.Button(self, label="OK")
        sizer_2.Add(cancel, 0, wx.ALL, 3)
        sizer_2.AddStretchSpacer(1)
        sizer_2.Add(self.ok, 0, wx.ALL, 3)
        sizer_1.Add(sizer_2, 0, wx.EXPAND|wx.ALL, 3)

        self.SetSizerAndFit(sizer_1)

        q_ctrl.Bind(EVT_QUANTITY_CHANGED, self._qty_changed)
        q_ctrl.Bind(EVT_UNIT_DEFINED, self._new_unit_defined)
        self.ok.Bind(wx.EVT_BUTTON, lambda e: self.EndModal(wx.ID_OK))
        cancel.Bind(wx.EVT_BUTTON, lambda e: self.EndModal(wx.ID_CANCEL))
        self._value = value
        self._item = item

    @property
    def value(self) -> Quantity:
        return self._value

    def _qty_changed(self, event):
        self.ok.Enable(event.in_bounds)
        self._value = event.new_value
        if event.in_bounds:
            self.ok.SetFocus()

    @staticmethod
    def _new_unit_defined(event):
        pub.sendMessage(NEW_UNIT_DEFINED, unit=event.new_unit)



from collections.abc import Collection

import wx
from wx.lib.newevent import NewCommandEvent
from pint import (Unit, DimensionalityError, DefinitionSyntaxError,
                  UndefinedUnitError)
from pint.registry import Quantity
from ..utils import fmt_unit, get_unit_registry
from .colors import ERROR_RED, LIGHT_GREY


# create a new custom event class
_QuantityChangedEventBase, EVT_QUANTITY_CHANGED = NewCommandEvent()
NewUnitDefinedEvent, EVT_UNIT_DEFINED = NewCommandEvent()


class QuantityChangedEvent(_QuantityChangedEventBase):
    MAGNITUDE = 0
    UNIT = 1
    ENTER = 2
    FOCUS_LOST = 4

    def __init__(self, parent, value, in_bounds, base_type, evt_type):
        super().__init__(id=parent.GetId())
        self.SetEventObject(parent)
        self.new_value = value
        self.in_bounds = in_bounds

        self.type = base_type
        if evt_type == wx.EVT_TEXT_ENTER.typeId:
            self.type |= QuantityChangedEvent.ENTER
        elif evt_type == wx.EVT_KILL_FOCUS.typeId:
            self.type |= QuantityChangedEvent.FOCUS_LOST

    def enter_pressed(self):
        return bool(self.type & self.ENTER)

    def focus_lost(self):
        return bool(self.type & self.FOCUS_LOST)


class QuantityCtrl(wx.Window):
    ERROR_SHOW_DURATION = 1500  # ms

    def __init__(self, parent, value: Quantity, units: Collection[str],
                 min_value: float = None, max_value: float = None):
        super().__init__(parent)
        value_str = f"{value.m:.7g}"
        unit_str = fmt_unit(value.u)
        self.value = value
        qty_cls = get_unit_registry().Quantity
        self.min = None if min_value is None else qty_cls(min_value, value.u)
        self.max = None if max_value is None else qty_cls(max_value, value.u)
        self.units = set(units)

        self.magnitude_ctrl = wx.TextCtrl(
            self, value=value_str, style=wx.TE_PROCESS_ENTER)
        self.magnitude_ctrl.SetMinSize(self.magnitude_ctrl.GetMinSize()
                                       + wx.Size(120, 0))
        self.unit_ctrl = wx.ComboBox(
            self, value=unit_str, choices=list(self.units),
            style=wx.CB_DROPDOWN | wx.CB_SORT | wx.TE_PROCESS_ENTER)

        self.unit_ctrl.SetMinSize(self.unit_ctrl.GetMinSize()
                                  + wx.Size(120, 0))
        self.link_ctrl = wx.ToggleButton(
            self, label="\N{LINK SYMBOL}", style=wx.BU_EXACTFIT)
        self.link_ctrl.SetValue(True)

        # status
        font = wx.Font(8, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL,
                       wx.FONTWEIGHT_NORMAL)
        self.status = wx.StaticText(self, label="")
        self.status.SetFont(font)

        sizer_1 = wx.BoxSizer(wx.VERTICAL)
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(self.magnitude_ctrl, 1, wx.EXPAND | wx.ALL, 3)
        sizer.Add(self.unit_ctrl, 0, wx.EXPAND | wx.ALL, 3)
        sizer.Add(self.link_ctrl, 0, wx.ALIGN_CENTER_VERTICAL, 3)
        sizer_1.Add(sizer, 0, wx.EXPAND)
        sizer_1.Add(self.status, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 2)
        self.SetSizerAndFit(sizer_1)
        self.SetMinSize(self.GetSize())

        self.unit_ctrl.Bind(wx.EVT_TEXT_ENTER, self._on_unit_changed)
        self.unit_ctrl.Bind(wx.EVT_COMBOBOX, self._on_unit_changed)
        self.magnitude_ctrl.Bind(wx.EVT_TEXT_ENTER, self._on_magnitude_changed)
        self.magnitude_ctrl.Bind(wx.EVT_KILL_FOCUS, self._on_magnitude_changed)

    def _on_unit_changed(self, event):
        def show_error(message):
            self.status.SetLabelText(message)
            self.status.SetForegroundColour(ERROR_RED)
            self.unit_ctrl.SetBackgroundColour(ERROR_RED)
            wx.CallLater(self.ERROR_SHOW_DURATION, reset)

        def reset():
            self.unit_ctrl.SetBackgroundColour(orig_bg)
            self.unit_ctrl.SetValue(fmt_unit(self.value.u))
            self.status.SetLabelText("You can do this!")
            self.status.SetForegroundColour(LIGHT_GREY)

        event.Skip()
        orig_bg = self.unit_ctrl.GetBackgroundColour()
        candidate = event.GetString()
        try:
            new_value = self.value.to(candidate)
        except DimensionalityError:
            show_error("Incompatible dimension")
            return
        except (DefinitionSyntaxError, AssertionError):
            show_error("Unit syntax error")
            return
        except UndefinedUnitError:
            show_error("Undefined unit")
            return

        # add unit to set if it doesn't exist yet
        #  maybe even fire an event to add it to global list
        if (cand_fmt := fmt_unit(Unit(candidate))) not in self.units:
            self.units.add(cand_fmt)
            self.unit_ctrl.Append(cand_fmt)
            evt = NewUnitDefinedEvent(self.GetId())
            evt.SetEventObject(self)
            evt.new_unit = cand_fmt
            wx.PostEvent(self, evt)

        self.unit_ctrl.SetValue(cand_fmt)

        # convert value of magnitude if link is active
        if self.link_ctrl.GetValue():
            self.value = new_value
            self.magnitude_ctrl.SetValue(f"{self.value.m:.7g}")
        else:
            qty_cls = get_unit_registry().Quantity
            self.value = qty_cls(self.value.m, cand_fmt)

        self._fire_change_event(QuantityChangedEvent.UNIT, event.GetEventType())

    def _on_magnitude_changed(self, event):
        def show_error(message):
            self.status.SetLabelText(message)
            self.status.SetForegroundColour(ERROR_RED)
            self.magnitude_ctrl.SetBackgroundColour(ERROR_RED)
            wx.CallLater(self.ERROR_SHOW_DURATION, reset)

        def reset():
            self.magnitude_ctrl.SetBackgroundColour(orig_bg)
            self.magnitude_ctrl.SetValue(f"{self.value.m:.7g}")
            self.status.SetForegroundColour(LIGHT_GREY)
            self.status.SetLabelText("You can do this!")

        event.Skip()
        orig_bg = self.magnitude_ctrl.GetBackgroundColour()
        try:
            new_value = float(self.magnitude_ctrl.GetValue())
        except ValueError:
            show_error("Invalid number format")
            return

        qty_cls = get_unit_registry().Quantity
        self.value = qty_cls(new_value, self.value.u)
        self._fire_change_event(QuantityChangedEvent.MAGNITUDE,
                                event.GetEventType())

    def _fire_change_event(self, base_type, evt_type):
        evt = QuantityChangedEvent(
            self, self.value, self._check_bounds(), base_type, evt_type)
        wx.PostEvent(self, evt)

    def _check_bounds(self) -> bool:
        value, v_min, v_max = self.value, self.min, self.max
        in_bounds = True if v_min is None else (v_min <= value)
        in_bounds &= True if v_max is None else (value <= v_max)
        if in_bounds:
            self.status.SetLabelText("")
            self.status.SetForegroundColour(LIGHT_GREY)
        else:
            ## oo looks way better than \N{INFINITY}, which is too small.
            min_fmt = "-oo" if v_min is None else f"{v_min.to(value.u):.6g~P}"
            max_fmt = "oo" if v_max is None else f"{v_max.to(value.u):.6g~P}"
            msg = f"Value out of bounds: [{min_fmt}, {max_fmt}]"
            self.status.SetLabelText(msg)
            self.status.SetForegroundColour(ERROR_RED)
        return in_bounds



from collections.abc import Mapping
from typing import Any

import wx
from wx.dataview import (
    PyDataViewModel, DataViewItem, NullDataViewItem, DataViewCtrl,
    DV_HORIZ_RULES, DV_ROW_LINES, DV_VERT_RULES, DV_MULTIPLE,
    EVT_DATAVIEW_ITEM_ACTIVATED, EVT_DATAVIEW_ITEM_CONTEXT_MENU,
    EVT_DATAVIEW_ITEM_COLLAPSED, EVT_DATAVIEW_ITEM_EXPANDED)
from pubsub import pub
from pint.registry import Quantity

from ..utils import fmt_unit, PathFilter, DataStructure
from ..events import RESULT_UNIT_CLICKED, NEW_UNIT_DEFINED, RESULT_UNIT_CHANGED
from .auxiliary import PopupBase


class ResultViewModel(PyDataViewModel):
    def __init__(self):
        super().__init__()
        self.data = DataStructure()
        self._filtered_data = self.data
        self._filter_term = ""
        self._items = {}

    def set_data(self, data: DataStructure):
        self.data = data
        self.apply_filter(self._filter_term)
        self._items = {}
        self.Cleared()

    def ObjectToItem(self, path):
        # required so that same items have same ids
        if path not in self._items:
            item = super().ObjectToItem(path)
            self._items[path] = item
        return self._items[path]

    def IsContainer(self, item: DataViewItem) -> bool:
        if not item:
            return True
        path = self.ItemToObject(item)
        return isinstance(self._filtered_data.get(path), Mapping)

    def GetParent(self, item: DataViewItem) -> DataViewItem:
        if not item:
            return NullDataViewItem
        path = self.ItemToObject(item)
        if len(path) == 1:
            return NullDataViewItem
        result = self.ObjectToItem(path[:-1])
        return result

    def HasValue(self, item: DataViewItem, col: int):
        path = self.ItemToObject(item)
        return col == 0 or not self.IsContainer(item)

    def GetChildren(self, item: DataViewItem,
                    children: list[DataViewItem]) -> int:
        path = self.ItemToObject(item) if item else ()
        if isinstance(data := self._filtered_data.get(path), Mapping):
            for k in data:
                children.append(self.ObjectToItem(path + (k, )))
        return len(children)

    def GetValue(self, item: DataViewItem, col: int) -> str:
        path = self.ItemToObject(item)
        if col == 0:
            return path[-1]
        value = self._filtered_data.get(path)
        if isinstance(value, Mapping):
            return ""
        return f"{value.m:.6g}" if col == 1 else f"{value.u:~P}"

    def all_values_changed(self):
        def dive(item: DataViewItem):
            if self.IsContainer(item):
                self.GetChildren(item, children := [])
                for c in children:
                    dive(c)
            else:
                self.ValueChanged(item, 1)
                self.ValueChanged(item, 2)
        dive(NullDataViewItem)

    def apply_filter(self, term: str):
        def dive(d: DataStructure, path: tuple[str, ...]):
            if isinstance(d, Mapping):
                result = {k: dive(v, path + (k, )) for k, v in d.items()}
                result =  {k: v for k, v in result.items() if v is not None}
                return result if result else None
            else:
                return d if filter_.matches(path) else None

        self._items = {}
        self._filter_term = term
        filter_ = PathFilter(term)
        data = dive(self.data, ())
        self._filtered_data = DataStructure({} if data is None else data)
        self.Cleared()


class UnitPopup(PopupBase):
    def __init__(self, parent: wx.Window, active_unit: str,
                 units: list[str], callback, rect):
        self._units = units
        super().__init__(parent, rect, active_unit, callback)
        self.bind_ctrl(wx.EVT_COMBOBOX)
        self.bind_ctrl(wx.EVT_TEXT_ENTER)

    def create_ctrl(self, parent, initial_value):
        style = wx.CB_DROPDOWN | wx.CB_SORT | wx.TE_PROCESS_ENTER
        return wx.ComboBox(parent, value=initial_value,
                           choices=self._units, style=style)
    def collect_result(self, event: wx.CommandEvent, ctrl):
        return event.GetString()


class ResultDataViewCtrl(DataViewCtrl):
    def __init__(self, parent: wx.Window):
        style = (wx.BORDER_THEME | DV_ROW_LINES | DV_HORIZ_RULES |
                 DV_VERT_RULES | DV_MULTIPLE)
        super().__init__(parent, style=style)

        self.SetMinSize(wx.Size(500, 400))

        self.Bind(EVT_DATAVIEW_ITEM_ACTIVATED, self._on_item_activated)
        self.Bind(EVT_DATAVIEW_ITEM_CONTEXT_MENU, self._on_right_click)
        self.Bind(EVT_DATAVIEW_ITEM_COLLAPSED, lambda e: self._get_expanded())
        self.Bind(EVT_DATAVIEW_ITEM_EXPANDED, lambda e: self._get_expanded())
        self.model = ResultViewModel()
        self.AssociateModel(self.model)
        self.model.DecRef()
        self.AppendTextColumn("Path", 0, width=250)
        self.AppendTextColumn("Magnitude", 1, width=150)
        self.AppendTextColumn("Unit", 2, width=100)
        self._popup_ids = None
        self._mouse_event = None
        self._expanded = set() # store all items (or ids) that are expanded
        # whether to react to collapsed and expanded events right now
        self._record_expanded = True
        self._get_expanded()

    def _on_right_click(self, event):
        self._mouse_event = event
        col = event.GetColumn()
        has_value = self.model.HasValue(event.GetItem(), 2)
        if col != 2 or not has_value:
            return

        menu_items = {"Apply to all": self._on_apply_to_all}

        if self._popup_ids is None:
            ids = {}
            for n, cb in menu_items.items():
                ids[n] = wx.NewIdRef()
                self.Bind(wx.EVT_MENU, cb, id=ids[n])
            self._popup_ids = ids

        menu = wx.Menu()
        for n, ref_id in self._popup_ids.items():
            menu.Append(ref_id, n)
        self.PopupMenu(menu)
        menu.Destroy()
        self._mouse_event = None

    def _on_apply_to_all(self, event):
        ev = self._mouse_event
        item = ev.GetItem()
        col = ev.GetColumn()
        model = self.model
        path = model.ItemToObject(item)
        value = model.data.get(path)
        if isinstance(value, Quantity):
            model.data.convert_all_possible_to(value.u)
        model.all_values_changed()
        pub.sendMessage(RESULT_UNIT_CHANGED)

    def _get_expanded(self):
        def dive(item: DataViewItem):
            if not (model.IsContainer(item) and
                    self.IsExpanded(item) or not item):
                return
            if item:
                expanded.add(model.ItemToObject(item))
            model.GetChildren(item, children := [])
            for c in children:
                dive(c)

        if not self._record_expanded:
            return
        expanded = set()
        model = self.model
        dive(NullDataViewItem)
        self._expanded = expanded

    def _apply_expanded(self):
        def dive(item: DataViewItem):
            if not model.IsContainer(item):
                return
            expand = True
            if item:
                path = model.ItemToObject(item)
                expand = path in self._expanded
                if expand:
                    expanded.add(path)
            if not expand:
                return
            self.Expand(item)
            model.GetChildren(item, children := [])
            for c in children:
                dive(c)

        expanded = set()
        model = self.model
        dive(NullDataViewItem)
        self._expanded = expanded

    def on_collapse_all(self, event):
        def collapse(item):
            self.model.GetChildren(item, children := [])
            for c in children:
                collapse(c)
                self.Collapse(c)
        collapse(NullDataViewItem)

    def on_expand_all(self, event):
        def expand(item):
            self.model.GetChildren(item, children := [])
            for c in children:
                self.ExpandChildren(c)
        expand(NullDataViewItem)

    def _on_item_activated(self, event):
        if event.GetColumn() != 2:
            return
        item = event.GetItem()
        path = self.model.ItemToObject(item)
        q = self.model.data.get(path)
        if isinstance(q, Quantity):
            pub.sendMessage(RESULT_UNIT_CLICKED, item=item, value=q)

    def change_unit(self, item, units):
        def commit(new_unit):
            try:
                self.model.data.set(path, value.to(new_unit))
                # should fire event to refresh results on canvas
            except Exception as e:  # this can be a lot of different ones
                msg = f"Invalid unit: {e}"
                style = wx.ICON_ERROR| wx.OK
                title = "Unit of measurement error"
                wx.MessageDialog(self, msg, title, style=style).ShowModal()
                return False
            self.model.ItemChanged(item)
            if new_unit not in units:
                pub.sendMessage(NEW_UNIT_DEFINED, unit=new_unit)
            pub.sendMessage(RESULT_UNIT_CHANGED)
            return True

        col = self.GetColumn(2)
        path = self.model.ItemToObject(item)
        value = self.model.data.get(path)
        units = list(units | {active_unit := fmt_unit(value.u)})

        # position and size
        rect = self.GetItemRect(item, col)
        rect.SetPosition(rect.GetPosition() + self.GetPosition())

        popup = UnitPopup(self.GetParent(), active_unit,
                          units, commit, rect)
        wx.CallAfter(popup.ShowModal)

    def set_data(self, data: DataStructure):
        self._record_expanded = False
        self.model.set_data(data)
        self._apply_expanded()
        self._record_expanded = True

    def on_search(self, term: str):
        self._record_expanded = False
        self.model.apply_filter(term)
        self._apply_expanded()
        self._record_expanded = True


class ResultView(wx.Frame):
    def __init__(self, parent: wx.Window):
        style = wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER | wx.MAXIMIZE_BOX
        super().__init__(parent, id=wx.ID_ANY, title="Results", style=style)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.search = wx.SearchCtrl(self, style=wx.TE_PROCESS_ENTER)
        self.search.ShowCancelButton(True)
        sizer.Add(self.search, 0, wx.EXPAND | wx.ALL, 3)

        self.view_ctrl = ResultDataViewCtrl(self)
        sizer.Add(self.view_ctrl, 1, wx.EXPAND | wx.ALL, 3)
        self.SetSizerAndFit(sizer)

        menu_bar = wx.MenuBar()

        view_menu = wx.Menu()
        item = view_menu.Append(wx.ID_ANY, "E&xpand all\tCTRL+e",
                                  "Expand all nodes")
        self.Bind(wx.EVT_MENU, self.view_ctrl.on_expand_all, item)
        item = view_menu.Append(wx.ID_ANY, "&Collapse all\tCTRL+w",
                                "Collapse all nodes")
        self.Bind(wx.EVT_MENU, self.view_ctrl.on_collapse_all, item)
        menu_bar.Append(view_menu, "&View")

        self.SetMenuBar(menu_bar)
        self.Bind(wx.EVT_CLOSE, lambda e: self.Show(False))
        # for e in (wx.EVT_SEARCH, wx.EVT_SEARCH_CANCEL,
        #           wx.EVT_KILL_FOCUS, wx.EVT_TEXT_ENTER):
        self.search.Bind(wx.EVT_SEARCH, self._on_search)

    def _on_search(self, event):
        self.view_ctrl.on_search(self.search.GetValue())



from collections.abc import MutableMapping, Set
from typing import Any
from pint.registry import Quantity

import wx
from pubsub.pub import sendMessage

from ..config import Configuration
from ..events import SHOW_PARAMETER_IN_CANVAS
from ..utils import DataStructure, fmt_unit, get_unit_registry
from ..models.tooltip import TooltipInfo

from .parameter import ParameterDialog
from .colors import INPUT_BLUE, BLACK, ERROR_RED, LIGHT_GREY
from .auxiliary import APoint, ASize

_TOOLTIP_CHECK_INTERVAL = 200  # ms
_TOOLTIP_DURATION = 5  # times check interval -> duration after hover
PROP_STUB = """
  - path: []
    uom: ""
    pos: [{x:d}, {y:d}]
    fmt: "{{:.2f~P}}"
"""[1:]

class Canvas(wx.ScrolledWindow):
    RESULT_VALID = 0
    RESULT_INVALID = 1
    RESULT_ERROR = 2

    def __init__(self, parent: wx.Window, config: Configuration):
        super().__init__(parent)
        self.config = config
        self._result_labels = []
        self._results_mode = False
        self._parameter_labels = []
        self._dirty_parameter_ids = set()
        self._tooltip = TooltipInfo()

        # configure bg picture
        self.background = config.get_image(config["bg_picture_name"])
        self.bg_size = config.get_image_size(self.background)

        panel = wx.Panel(self)
        panel.SetSize(self.bg_size)
        panel.SetMinSize(self.bg_size)
        panel.SetMaxSize(self.bg_size)
        panel.SetBackgroundStyle(wx.BG_STYLE_PAINT)
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(panel, 1, wx.EXPAND, 0)
        self.SetSizer(sizer)
        self.panel = panel

        self.SetScrollRate(50, 50)

        panel.Bind(wx.EVT_PAINT, self._on_paint)
        self.Bind(wx.EVT_SCROLLWIN, self._on_scroll)
        self.Bind(wx.EVT_SCROLL_THUMBRELEASE, self._on_scroll)
        self.Bind(wx.EVT_SCROLL_CHANGED, self._on_scroll)

        self.Bind(wx.EVT_MOUSEWHEEL, self._on_mousewheel)
        panel.Bind(wx.EVT_LEFT_DOWN, self._on_left_click)
        def set_size():
            self.SetVirtualSize(self.bg_size)

        wx.CallLater(50, set_size)

        self._tool_tip_timer = wx.Timer()
        self._tool_tip_timer.Start(_TOOLTIP_CHECK_INTERVAL)

        parent.Bind(wx.EVT_CLOSE, self._on_close)
        self._tool_tip_timer.Bind(wx.EVT_TIMER, self._on_check_tooltip)

    def _on_scroll(self, event):
        self.panel.Update()
        self.panel.Refresh()
        self.Refresh()
        event.Skip()

    def _on_close(self, event):
        # stop timer, or else it can still fire an event and trigger access
        # to destroyed canvas object
        self._tool_tip_timer.Stop()
        event.Skip()

    def _on_check_tooltip(self, event):
        def process(e):
            if not tooltip.is_for(e):
                tooltip.set_item(e)
            else:
                if tooltip.counter <= 0:
                    tooltip.pos = pos_canvas
                    tooltip.pos_panel = pos
                tooltip.counter = _TOOLTIP_DURATION  # reset counter
            self.Refresh()  # tooltip visibility changed

        tooltip = self._tooltip
        tooltip.counter -= 1
        if tooltip.counter == 0:
            self.Refresh()  # tooltip no longer visible

        pos = self.ScreenToClient(wx.GetMousePosition())
        pos_canvas = self._get_pos(pos)
        for item in self._result_labels + self._parameter_labels:
            try:
                if item["hitbox"].Contains(pos_canvas):
                    process(item)
                    return
            except KeyError:
                pass
        # self.panel.Refresh(False)
        # self.panel.Update()

    def _on_left_click(self, event: wx.MouseEvent):
        pos = event.GetPosition()

        # copy parameter / property definition stub to clipboard if pressed
        # with <ctrl>-key.

        if event.ControlDown():
            text = PROP_STUB.format(x=pos[0], y=pos[1])
            data_obj = wx.TextDataObject(text)
            print(text)

            if wx.TheClipboard.Open():
                wx.TheClipboard.SetData(data_obj)
                wx.TheClipboard.Close()
            else:
                print("Clipboard not available!")
            return

        for item in self._result_labels:
            if item["hitbox"].Contains(pos):
                print(item["label"])
                # TODO: process e.g. by asking for values for other scenarios
                #  .. fire this as an event to controller, including mouse
                #     position and path. Controller then calls back to show
                #     tooltip with values for each scenario.

        for item in self._parameter_labels:
            if item["hitbox"].Contains(pos):
                sendMessage(SHOW_PARAMETER_IN_CANVAS, item=item)


    def _on_mousewheel(self, event: wx.MouseEvent):
        if event.ShiftDown(): # horizontal scroll
            delta = event.GetWheelRotation()
            # positive delta = wheel up -> scroll left (negative)
            units = self.GetScrollPixelsPerUnit()[0]
            x, y = self.GetViewStart()
            self.Scroll(x - int(delta / event.GetWheelDelta()), y)
        else:  # default vertical scroll
            event.Skip()

    def draw_content(self, gc: wx.GraphicsContext):
        """Draw the SVG and all overlays onto the given GraphicsContext."""
        def draw_labels(which, font_clean, color_clean,
                        font_dirty=None, color_dirty=None):
            font_dirty = font_clean if font_dirty is None else font_dirty
            color_dirty = color_clean if color_dirty is None else color_dirty
            for e in which:
                dirty = e["id"] in self._dirty_parameter_ids
                gc.SetFont(font_dirty if dirty else font_clean,
                           color_dirty if dirty else color_clean)
                gc.DrawText(e["label"], *e["pos"])
                if "hitbox" not in e:
                    extent = gc.GetFullTextExtent(e["label"])
                    size = wx.Size(int(extent[0]), int(extent[1]))
                    e["hitbox"] = wx.Rect(wx.Point(*e["pos"]), size)

        # fill background explicitly
        bg = wx.Colour(self.config.get("bg_color", "white"))
        gc.SetBrush(gc.CreateBrush(wx.Brush(bg)))
        gc.SetPen(wx.TRANSPARENT_PEN)
        gc.DrawRectangle(0, 0, *self.panel.GetSize())
        # load background picture
        self.background.render_to_gc(gc, size=self.bg_size)

        # draw calculated properties
        font_size = self.config.get("font_size", 12)
        font = wx.Font(
            font_size, wx.FONTFAMILY_DEFAULT,
            wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, False)
        color = {self.RESULT_INVALID: LIGHT_GREY,
                 self.RESULT_VALID: BLACK,
                 self.RESULT_ERROR: ERROR_RED}[self._results_mode]

        draw_labels(self._result_labels, font, color)
        draw_labels(self._parameter_labels, font, INPUT_BLUE, font.Bold())

        self._draw_tooltip(gc)

    def _draw_tooltip(self, gc: wx.GraphicsContext):
        unit_cls = get_unit_registry().Unit
        tooltip = self._tooltip
        if tooltip.counter <= 0:
            return
        item = tooltip.item
        pos = APoint.from_point(tooltip.pos)
        try:
            tip = item["tip"]
        except KeyError:
            uom = fmt_unit(unit_cls(item["uom"]))
            tip = f"{item['name']} [{uom}]"
        tip = f"{tip}\n{item['label']}"
        extent = gc.GetFullTextExtent(tip)
        size = ASize(int(extent[0]) + 8, int(extent[1]) + 4)
        pos -= size // 2
        p_pos = APoint.from_point(tooltip.pos_panel)

        # make sure that we stay on the canvas
        rect = [p_pos - size // 2, p_pos + size // 2]
        canvas_size = self.GetSize()
        if rect[0].x < 0:
            pos.x -= rect[0].x
        if rect[0].y < 0:
            pos.y -= rect[0].y
        if (dx:= rect[1].x - canvas_size.x) > 0:
            pos.x -= dx
        if (dy:= rect[1].y - canvas_size.y) > 0:
            pos.y -= dy

        gc.SetPen(wx.Pen(INPUT_BLUE))
        gc.SetBrush(wx.Brush(LIGHT_GREY))
        gc.DrawRoundedRectangle(pos[0], pos[1], size[0], size[1], 4)
        font_size = self.config.get("font_size", 12)
        font = wx.Font(
            font_size, wx.FONTFAMILY_DEFAULT,
            wx.FONTSTYLE_SLANT, wx.FONTWEIGHT_NORMAL, False)
        gc.SetFont(font, INPUT_BLUE)
        gc.DrawText(tip, pos[0] + 4, pos[1] + 2)

    def _on_paint(self, event):
        dc = wx.AutoBufferedPaintDC(self.panel)
        dc.Clear()
        gc = wx.GraphicsContext.Create(dc)
        self.draw_content(gc)

    def _get_pos(self, view_pos: wx.Point) -> wx.Point:
        """The position on the virtual canvas is the view position added to
        the view start, which is wisely expressed not in pixels, but in
        scroll pixels per unit - why make it easy?"""
        vx, vy = self.GetViewStart()
        sx, sy = self.GetScrollPixelsPerUnit()
        return wx.Point(vx * sx + view_pos[0], vy * sy + view_pos[1])

    # ---- Export method ----
    def save_as_png(self, path: str):
        w, h = self.panel.GetSize()
        bmp = wx.Bitmap(w, h)
        mdc = wx.MemoryDC(bmp)
        mdc.Clear()
        self.panel.PrepareDC(mdc)

        gc = wx.GraphicsContext.Create(mdc)
        self.draw_content(gc)

        del gc
        mdc.SelectObject(wx.NullBitmap)
        bmp.SaveFile(path, wx.BITMAP_TYPE_PNG)

    def update_results(self, values: DataStructure, original: bool):
        mode = self.RESULT_VALID if original else self.RESULT_INVALID
        self.set_results_mode(mode)
        self._result_labels = self._entries(values, "results")
        if original:
            self._dirty_parameter_ids = set()

        self.Refresh()

    def update_parameters(self, values: DataStructure):
        self._parameter_labels = self._entries(values, "parameters")
        self.Refresh()

    def _entries(self, values: DataStructure, which: str):
        def e(item):
            try:
                q = values.get(item["path"])
            except KeyError:
                return None
            entry = {"label": item["fmt"].format(q.to(item["uom"])),
                     "id": ".".join(item["path"])}
            entry.update(item)
            if "name" not in entry:
                entry["name"] = entry["id"]
            return entry

        res = [e(item) for item in self.config[which]]
        return [r for r in res if r is not None]

    def show_parameter_dialog(self, item: MutableMapping[str, Any],
                              value: Quantity, units: Set[str]):
        dialog = ParameterDialog(self, item, value, units)
        if dialog.ShowModal() == wx.ID_OK:
            self.set_results_mode(self.RESULT_INVALID)
            self._dirty_parameter_ids.add(item["id"])
            return dialog.value
        return None

    def set_results_mode(self, mode: int):
        self._results_mode = mode


from wx import Colour

# can only create colors by RGB code, not by name, before wx.App is created
INPUT_BLUE = Colour(80, 80, 255)
LIGHT_GREY = Colour(200, 200, 200)
DARK_GREY = Colour(80, 80, 80)
ERROR_RED = Colour(255, 0, 0)
BLACK = Colour(0, 0, 0)
WARNING_ORANGE = Colour(220, 170, 0)




from importlib.metadata import version as _get_version
from .models.engine import CalculationEngine, CalculationFailed
from .utils import set_unit_registry, get_unit_registry
from .app import start_gui

__version__ = _get_version("wxfrog")



from importlib.resources import as_file, read_binary
from importlib.resources.abc import Traversable
from collections.abc import Sequence
from os import unlink
from yaml import safe_load
from pint.registry import Quantity
from tempfile import NamedTemporaryFile
import wx

from .views.image import SVGImageWrap, PNGImageWrap

CONFIG_FILENAME = "configuration.yml"


class Configuration(dict):
    def __init__(self, config_dir: Traversable):
        self.config_dir = config_dir
        with as_file(config_dir.joinpath(CONFIG_FILENAME)) as path:
            with open(path) as file:
                config = safe_load(file)
        super().__init__(config)

    def get_image(self, name):
        ending = name.split(".")[-1].lower()
        with as_file(self.config_dir.joinpath(name)) as path:
            with open(path, "rb") as file:
                if ending == "svg":
                    return SVGImageWrap(file)
                elif ending == "png":
                    return PNGImageWrap(file)
        raise ValueError(f"Unsupported file format `{ending}`.")

    def get_image_size(self, image) -> wx.Size:
        """If neither bg_picture_with nor bg_picture_height is defined, return
        the size of the original image.

        If one of these is defined, scale the other to match the aspect ratio.
        If both are defined, return the specified dimensions directly.
        """
        w_tg = self.get("bg_picture_width", -1)
        h_tg = self.get("bg_picture_height", -1)
        w, h = image.width, image.height
        if w_tg > 0 and h_tg > 0:
            return wx.Size(w_tg, h_tg)
        if w_tg <= 0 and h_tg <= 0:
            return wx.Size(w, h)
        if w_tg <= 0:
            return wx.Size(int(w * h_tg / h), h_tg)
        else:
            return wx.Size(w_tg, int(h * w_tg / w))

    def get_app_icon(self):
        try:
            name = self["app_icon"]
        except KeyError:
            return
        with as_file(self.config_dir.joinpath(name)) as path:
            with open(path, "rb") as file:
                data = file.read()

        tmp = NamedTemporaryFile(delete=False, suffix=".ico")
        tmp.write(data)
        tmp.close()
        icon = wx.Icon(tmp.name, wx.BITMAP_TYPE_ICO)
        unlink(tmp.name)
        return icon


class ConfigurationError:
    def __init__(self, path: Sequence[str], message: str):
        self.path = path
        self.message = message
        self.details = {}

    def _add_details(self, **kwargs: str):
        self.details.update(kwargs)


class ParameterNotFound(ConfigurationError):
    def __init__(self, path: Sequence[str]):
        super().__init__(path, "Parameter not found")


class UnitSyntaxError(ConfigurationError):
    def __init__(self, path: Sequence[str], unit: str):
        super().__init__(path, "Syntax error in unit of measurement")
        self._add_details(unit=unit)


class UndefinedUnit(ConfigurationError):
    def __init__(self, path: Sequence[str], unit: str):
        super().__init__(path, "Undefined unit of measurement")
        self._add_details(unit=unit)


class UnitConversionError(ConfigurationError):
    def __init__(self, path: Sequence[str], unit_model: str, unit_config: str):
        super().__init__(path, "Unit conversion error")
        self._add_details(unit_model=unit_model, unit_config=unit_config)


class OutOfBounds(ConfigurationError):
    def __init__(self, path: Sequence[str], value: Quantity,
                 bound: Quantity, max_bound: bool):
        super().__init__(path, "Parameter value out of bounds")
        self._add_details(value=f"{value:.6g~P}", bound=f"{bound:.6g~P}")
        self._add_details(which=("upper" if max_bound else "lower"))



from collections.abc import Set, Collection, MutableMapping
from threading import Thread
from copy import deepcopy
from math import nan

from pint import DimensionalityError, DefinitionSyntaxError, UndefinedUnitError
from pint.registry import Quantity
from pubsub import pub

from wxfrog.utils import (
    fmt_unit, ThreadedStringIO, get_unit_registry, DataStructure, Path)
from wxfrog.config import (
    Configuration, ConfigurationError, ParameterNotFound, UnitSyntaxError,
    UndefinedUnit, UnitConversionError, OutOfBounds)
from wxfrog.events import (
    INITIALIZATION_DONE, CALCULATION_DONE, CALCULATION_FAILED)
from .engine import CalculationEngine, CalculationFailed
from .casestudy import CaseStudy, ParameterSpec
from .scenarios import (Scenario, SCENARIO_DEFAULT, SCENARIO_CURRENT,
                        SCENARIO_CONVERGED)
from .html import HtmlTable


class Model:
    def __init__(self, engine: CalculationEngine, configuration: Configuration):
        self._configuration = configuration
        self._engine = engine
        self._out_stream = ThreadedStringIO()
        self._all_units = set()
        self._scenarios : MutableMapping[str, Scenario] = {}
        self._case_study = None
        self.file_path = None

    def initialise_engine(self):
        def f():
            self._engine.initialise(self._out_stream)
            pub.sendMessage(INITIALIZATION_DONE)

        Thread(target=f).start()

    def finalize_initialisation(self) -> Collection[ConfigurationError]:
        default_params = DataStructure(self._engine.get_default_parameters())
        snc = self._scenarios
        snc[SCENARIO_DEFAULT] = Scenario(default_params)
        snc[SCENARIO_CURRENT] = deepcopy(snc[SCENARIO_DEFAULT])
        u_cls = get_unit_registry().Unit
        self._all_units = {fmt_unit(u_cls(u))
                           for u in self._configuration.get("units", [])}
        errors = self._initialize_parameters(default_params)
        for r in self._configuration["results"]:
            self._all_units.add(fmt_unit(u_cls(r["uom"])))
        return errors

    @property
    def scenarios(self) -> MutableMapping[str, Scenario]:
        return self._scenarios

    @property
    def out_stream(self) -> ThreadedStringIO:
        return self._out_stream

    def run_engine(self):
        def f():
            param = scn.parameters
            try:
                # TODO: set initial values if available in scenario
                results = DataStructure(self._engine.calculate(param))
            except CalculationFailed as error:
                pub.sendMessage(CALCULATION_FAILED, message=str(error))
            else:
                scn.results = results
                self._scenarios[SCENARIO_CONVERGED] = scn
                pub.sendMessage(CALCULATION_DONE)

        scn = deepcopy(self._scenarios[SCENARIO_CURRENT])
        Thread(target=f, daemon=True).start()

    def compatible_units(self, value: Quantity) -> Set[str]:
        result = {u for u in self._all_units if value.is_compatible_with(u)}
        return result | {fmt_unit(value.u)}

    def register_unit(self, unit):
        unit_cls = get_unit_registry().Unit
        self._all_units.add(fmt_unit(unit_cls(unit)))

    def get_param_info(self, path):
        qty_cls = get_unit_registry().Quantity
        params = self.scenarios[SCENARIO_CURRENT].parameters
        min_, max_ = None, None
        name, uom = ".".join(path), ""
        for p in self._configuration["parameters"]:
            if tuple(p["path"]) == path:
                min_, max_ = p.get("min", None), p.get("max", None)
                name, uom = p.get("name", name), p["uom"]
                break
        value = params.get(path)
        min_bound, max_bound = 0.9 * value, 1.1 * value
        if min_ is not None:
            min_ = qty_cls(min_, uom)
            min_bound = max(min_, min_bound)
        if max_ is not None:
            max_ = qty_cls(max_, uom)
            max_bound = min(max_, max_bound)

        spec = ParameterSpec(path, min_bound, max_bound, name=name, num=5)
        units = self.compatible_units(value)
        return { "spec": spec, "min": min_, "max": max_, "units": units}

    def collect_stream_table(self, name: str) -> str:
        def get(d: DataStructure, p: Path, unit) -> float:
            try:
                return d.get(p).to(unit).m
            except KeyError:
                return nan

        data = self.scenarios[SCENARIO_CONVERGED].results
        if data is None:
            raise ValueError("No results available to export steam table")

        definition = self._configuration["tables"][name]
        columns, props_raw = definition["columns"], definition["properties"]
        streams = {k: [d[k] for d in columns] for k in ["label", "path"]}
        props = {k: [p[k] for p in props_raw]
                 for k in ["label", "uom", "path", "fmt"]}
        props["tolerance"] = [float(p.get("tolerance", 0)) for p in props_raw]
        table = HtmlTable(streams["label"], props["label"])
        table.label = name
        table.title = "Unnamed" if self.file_path is None else self.file_path
        table.add_row_unit_column(props["uom"])
        table.add_column_header_row([""] * len(streams["label"]))
        table.set_top_rect_headers([["Stream", ""], ["Property", "Unit"]])
        for k, (t, f) in enumerate(zip(props["tolerance"], props["fmt"])):
            table.set_threshold(k, t)
            table.set_nan(k, "")
            table.set_row_format(k, f)
        # set data
        table_data = [[get(data, s + p, u) for s in streams["path"]]
                      for p, u in zip(props["path"], props["uom"])]
        table.set_data(table_data)
        return table.render()

    def _initialize_parameters(
            self, param: DataStructure) -> Collection[ConfigurationError]:
        errors = []
        qty_cls = get_unit_registry().Quantity
        for item in self._configuration["parameters"]:
            path = item["path"]
            try:
                v = param.get(path)
            except KeyError:
                errors.append(ParameterNotFound(path))
                continue
            try:
                param.set(path, v.to(item["uom"]))
            except (DefinitionSyntaxError, AssertionError, ValueError):
                errors.append(UnitSyntaxError(path, item["uom"]))
                continue
            except UndefinedUnitError:
                errors.append(UndefinedUnit(path, item["uom"]))
                continue
            except DimensionalityError:
                errors.append(UnitConversionError(path, fmt_unit(v.u),
                                                  item["uom"]))
                continue
            i_min, i_max = item["min"], item["max"]
            if i_max is not None and v > (v_max := qty_cls(i_max, v.u)):
                errors.append(OutOfBounds(path, v, v_max, True))
            elif i_min is not None and v < (v_min := qty_cls(i_min, v.u)):
                errors.append(OutOfBounds(path, v, v_min, False))
            self._all_units.add(fmt_unit(v.u))
        return errors

    def assure_case_study(self):
        scn = self._scenarios[SCENARIO_CONVERGED]
        if self._case_study is None:
            self._case_study = CaseStudy(self._engine, scn, self._out_stream)
        return self._case_study

    def interrupt_case_study(self):
        self._case_study.interrupt()

    def collect_case_study_results(self, paths):
        return self._case_study.collect(self.file_path, paths)

    def serialize(self, case_study_param):
        case_study = self._case_study
        if case_study is not None:
            case_study = case_study.serialize(case_study_param)
        return {
            "units": list(self._all_units),
            "scenarios": {n: s.serialize() for n, s in self._scenarios.items()},
            "case_study": case_study
        }

    def deserialize(self, data):
        self._all_units = set(data["units"])
        self._scenarios = {n: Scenario.deserialize(d)
                           for n, d in data["scenarios"].items()}
        self._case_study, param = CaseStudy.deserialize(
            self._engine, self._out_stream, data["case_study"])
        return param




from copy import deepcopy
from importlib.resources.abc import Traversable
from pubsub import pub
from zipfile import ZipFile, ZIP_DEFLATED
from json import dumps, load
import wx

from .models.engine import CalculationEngine
from .models.model import Model
from .views.frame import FrogFrame
from .config import Configuration
from .events import *
from .utils import copy_html_to_clipboard
from wxfrog.models.scenarios import SCENARIO_CURRENT, SCENARIO_CONVERGED


class Controller:
    def __init__(self, config_directory: Traversable, model: CalculationEngine):
        self.configuration = Configuration(config_directory)
        self._running = True  # initially busy with initialization

        # event subscriptions
        events = {
            EXPORT_CANVAS_GFX: self._on_export_canvas_gfx,
            RUN_MODEL: self._on_model_run,
            SHOW_PARAMETER_IN_CANVAS: self._on_show_parameter,
            NEW_UNIT_DEFINED: self._on_new_unit_defined,
            INITIALIZATION_DONE: self._on_initialisation_done,
            CALCULATION_DONE: self._on_calculation_done,
            OPEN_SCENARIOS: self._on_open_scenarios,
            OPEN_FILE: self._on_open_file,
            SAFE_FILE: self._on_save_file,
            SAFE_FILE_AS: self._on_save_file_as,
            EXIT_APP: self._on_exit_app,
            RUN_CASE_STUDY: self._on_run_case_study,
            CALCULATION_FAILED: self._on_calculation_failed,
            COPY_SCENARIO: self._on_copy_scenario,
            RENAME_SCENARIO: self._on_rename_scenario,
            DELETE_SCENARIO: self._on_delete_scenario,
            OPEN_RESULTS: self._on_open_results,
            RESULT_UNIT_CLICKED: self._on_result_unit_clicked,
            RESULT_UNIT_CHANGED: self._update_results,
            CASE_STUDY_PARAMETER_SELECTED: self._on_case_study_param_sel,
            CASE_STUDY_RUN: self._on_case_study_run,
            CASE_STUDY_ENDED: self._on_case_study_ended,
            CASE_STUDY_INTERRUPT: self._on_interrupt_case_study,
            CASE_STUDY_PROPERTIES_SELECTED:
                self._on_case_study_properties_selected,
            COPY_STREAM_TABLE: self._on_copy_stream_table
        }

        for evt_id, callback in events.items():
            pub.subscribe(callback, evt_id)

        self.model = Model(model, self.configuration)
        self.frame = FrogFrame(self.configuration, self.model.out_stream)
        self.model.initialise_engine()
        self.frame.Show()

    def _on_copy_stream_table(self, name: str):
        html = self.model.collect_stream_table(name)
        copy_html_to_clipboard(html)

    def _on_case_study_run(self, specs):
        case_study = self.model.assure_case_study()
        case_study.set_parameters(specs)
        case_study.run()
        self.frame.run_menu_item.Enable(False)
        self.frame.case_study_menu_item.Enable(False)
        self.frame.case_studies.allow_run(False)
        self._running = True

    def _on_interrupt_case_study(self):
        self.model.interrupt_case_study()

    def _on_case_study_ended(self):
        self.frame.run_menu_item.Enable(True)
        self.frame.case_study_menu_item.Enable(True)
        self.frame.case_studies.allow_run(True)
        self._running = False

    def _on_case_study_properties_selected(self, paths):
        html = self.model.collect_case_study_results(paths)
        copy_html_to_clipboard(html)

    def _on_export_canvas_gfx(self):
        msg = "Save canvas as graphics"
        path = self.frame.show_file_dialog(msg, "PNG files", "png", save=True)
        if path is not None:
            self.frame.canvas.save_as_png(path)

    def _on_model_run(self):
        self.model.run_engine()
        self.frame.run_menu_item.Enable(False)
        self.frame.case_study_menu_item.Enable(False)
        self.frame.case_studies.allow_run(False)
        self.frame.copy_stream_table_menu_item.Enable(False)
        self._running = True

    def _on_calculation_done(self):
        # if parameters of converged are still the same as current, copy them.
        scn = self.model.scenarios
        if scn[SCENARIO_CURRENT].modified == scn[SCENARIO_CONVERGED].modified:
            scn[SCENARIO_CURRENT].results = scn[SCENARIO_CONVERGED].results
        self._update_results()
        self._update_scenarios()
        self.frame.run_menu_item.Enable()
        self.frame.case_study_menu_item.Enable()
        self.frame.case_studies.allow_run(True)
        self.frame.copy_stream_table_menu_item.Enable()
        self._running = False

    def _on_calculation_failed(self, message):
        self.frame.show_calculation_error(message)
        self.frame.run_menu_item.Enable()
        self.frame.case_study_menu_item.Enable()
        self.frame.case_studies.allow_run(True)
        self._running = False

    def _on_initialisation_done(self):
        errors = self.model.finalize_initialisation()
        if errors:
            wx.CallAfter(self.frame.show_config_error_dialog, errors)
        wx.CallAfter(self._update_parameters)
        self.frame.case_studies.switch_button_enable("add", True)
        self.frame.case_studies.allow_run(True)
        self.frame.run_menu_item.Enable()
        self.frame.case_study_menu_item.Enable()
        self._running = False
        if self.configuration.get("run_engine_on_start", False):
            self._on_model_run()

    def _on_open_results(self):
        self.frame.results.Show()

    def _on_open_scenarios(self):
        self._update_scenarios()
        self.frame.scenarios.Show()

    def _on_open_file(self):
        msg = "Save file"
        app_name = self.configuration["app_name"]
        ending = self.configuration["file_ending"]
        path = self.frame.show_file_dialog(msg, app_name, ending, save=False)
        if path is None:
            return

        with ZipFile(path, "r") as zip_file:
            with zip_file.open("data.json") as file:
                data = load(file)
        self.model.file_path = path
        self.frame.update_title(path)
        param = self.model.deserialize(data)
        self.frame.case_studies.list_ctrl.set_parameters(param)

        self._update_parameters()
        self._update_results()
        self._update_scenarios()

    def _on_save_file(self):
        path = self.model.file_path
        if path is None:
            self._on_save_file_as()
            return
        # hack to get parameters from case study from view
        # (I know, it breaks MVC, to be changed later)
        param = self.frame.case_studies.list_ctrl.get_parameters()
        data = self.model.serialize(param)
        json = dumps(data, indent=2, ensure_ascii=False)
        with ZipFile(path, "w", compression=ZIP_DEFLATED) as file:
            file.writestr("data.json", json)

    def _on_save_file_as(self):
        msg = "Save file"
        app_name = self.configuration["app_name"]
        ending = self.configuration["file_ending"]
        path = self.frame.show_file_dialog(msg, app_name, ending, save=True)
        if path is not None:
            self.model.file_path = path
            self._on_save_file()
            self.frame.update_title(path)

    def _on_exit_app(self, event):
        # may ask to save file
        event.Skip()

    def _on_run_case_study(self):
        scenario = self.model.scenarios[SCENARIO_CURRENT]
        self.frame.show_case_studies(scenario)

    def _on_show_parameter(self, item):
        scn_current = self.model.scenarios[SCENARIO_CURRENT]
        param = scn_current.parameters
        value = param.get(item["path"])
        units = self.model.compatible_units(value)
        new_value = self.frame.canvas.show_parameter_dialog(item, value, units)
        if new_value is None or new_value == value:
            return
        scn_current.set_param(item["path"], new_value)
        self.frame.canvas.update_parameters(param)
        if (self.configuration.get("run_engine_on_change", False)
                and not self._running):
            self._on_model_run()

    def _on_new_unit_defined(self, unit: str):
        self.model.register_unit(unit)

    def _on_copy_scenario(self, source: str, target: str):
        scenarios = self.model.scenarios
        scenarios[target] = deepcopy(scenarios[source])
        if target == SCENARIO_CURRENT:
            self._update_parameters()
            self._update_results()
        self._update_scenarios()

    def _on_rename_scenario(self, source: str, target: str):
        scenarios = self.model.scenarios
        scenarios[target] = scenarios[source]
        del scenarios[source]
        self._update_scenarios()

    def _on_delete_scenario(self, name: str):
        scenarios = self.model.scenarios
        del scenarios[name]
        self._update_scenarios()

    def _on_result_unit_clicked(self, item, value):
        units = self.model.compatible_units(value)
        self.frame.results.view_ctrl.change_unit(item, units)

    def _on_case_study_param_sel(self, path):
        param_info = self.model.get_param_info(path)
        self.frame.case_studies.list_ctrl.add(param_info)

    # non-event standard workflows, triggered by events

    def _update_parameters(self):
        self.frame.canvas.update_parameters(
            self.model.scenarios[SCENARIO_CURRENT].parameters)

    def _update_results(self):
        scn = self.model.scenarios
        current = scn[SCENARIO_CURRENT].has_results()
        which = SCENARIO_CURRENT if current else SCENARIO_CONVERGED
        self.frame.canvas.update_results(scn[which].results, current)
        self.frame.results.view_ctrl.set_data(scn[which].results)

    def _update_scenarios(self):
        scn = {name: (s.has_results(), s.mod_local_time())
               for name, s in self.model.scenarios.items()}
        self.frame.scenarios.update(scn)




from typing import Optional
from collections.abc import Sequence
from dataclasses import dataclass, KW_ONLY, field
from threading import Thread, Lock
from time import sleep
from itertools import product
from copy import deepcopy
from math import log, ceil

from pubsub import pub
from pint.registry import Quantity  # actual type

from wxfrog.utils import DataStructure, get_unit_registry
from .engine import CalculationEngine, CalculationFailed
from .html import HtmlTable
from .scenarios import Scenario
from ..events import CALCULATION_FAILED, CASE_STUDY_ENDED, CASE_STUDY_PROGRESS
from ..utils import Path


@dataclass
class ParameterSpec:
    path: Path
    min: Quantity
    max: Quantity
    _: KW_ONLY
    name: str = None
    incr: Quantity|float = None
    num: int = None
    log: bool = False
    data: Sequence[Quantity] = field(init=False, repr=False, compare=False)
    num_spec: bool = field(init=False, repr=False, compare=False)

    def __post_init__(self):
        if self.name is None:
            self.name = ".".join(self.path)
        self.num_spec = self.incr is None
        if self.log:
            self._post_init_log()
        else:
            self._post_init_linear()

    def _post_init_linear(self):
        num, incr = self.num, self.incr
        min_, max_ = self.min, self.max
        interval = max_ - min_
        if self.num_spec:
            if num is None:
                num = 11
            incr = interval / (num - 1) if num > 1 else min_ * 0
        else:
            assert num is None, "Only either num or incr can be specified"
            if incr / interval < 0:
                incr = -incr
            num = 1 + ceil(interval / incr)
        self.data = [min_ + i * incr for i in range(num - 1)] + [max_]
        self.num, self.incr = num, incr

    def serialize(self):
        return {
            "path": self.path,
            "min": f"{self.min:.14g~}",
            "max": f"{self.max:.14g~}",
            "name": self.name,
            "num": self.num,
            "log": self.log,
            "incr": self.incr if self.log else f"{self.incr:.14g~}",
            "num_spec": self.num_spec
        }

    @classmethod
    def deserialize(cls, data):
        qty_cls = get_unit_registry().Quantity
        is_log = data["log"]
        if data["num_spec"]:
            num, incr = data["num"], None
        else:
            num = None
            incr = float(data["incr"]) if is_log else qty_cls(data["incr"])

        return cls(
            tuple(data["path"]), qty_cls(data["min"]), qty_cls(data["max"]),
            name=data["name"], log=is_log, incr=incr, num=num
        )

    def _post_init_log(self):
        num, incr = self.num, self.incr
        min_, max_ = self.min, self.max
        ratio = (max_ / min_).to("")
        if self.num_spec:
            if num is None:
                num = 11
            incr = float(ratio ** (1 / (num - 1))) if num > 1 else 1
        else:
            assert num is None, "Only either num or incr can be specified"
        if (incr - 1) * (ratio - 1) < 0:
            incr = 1 / incr
        num = 1 + ceil(log(ratio) / log(incr))
        self.data = [min_ * incr ** i for i in range(num - 1)] + [max_]
        self.num, self.incr = num, incr


class CaseStudyResults:
    def __init__(self, param_paths: Sequence[Path],
                 result_paths: Sequence[Path]):
        self.param_columns = param_paths
        self.result_columns = result_paths
        self.params: list[Sequence[Quantity]] = []
        self.results: list[Sequence[Quantity]] = []

    def add_result(self, params: DataStructure, results: DataStructure):
        self.params.append([params.get(p) for p in self.param_columns])
        self.results.append([results.get(p) for p in self.result_columns])

    def collect(self, filename: str, paths: Sequence[Path]) -> str:
        mask = self._filter_properties(paths)
        param_names = [".".join(p) for p in self.param_columns]
        param_units = [str(p.u) for p in self.params[0]]
        prop_names = [".".join(p)
                      for p, m in zip(self.result_columns, mask) if m]
        prop_units = [str(p.u) for p, m in zip(self.results[0], mask) if m]
        row_labels = list(map(str, range(1, len(self.results) + 1)))
        table = HtmlTable(param_names + prop_names, row_labels)
        table.label = "Case study"
        table.title = "Unnamed" if filename is None else filename
        table.add_column_unit_row(param_units + prop_units)
        table.add_vertical_line(len(param_names) - 1)
        table.set_top_rect_headers([["Parameter"], ["Unit"]])

        data = [
            [pa.m for pa in param_row] +
            [pr.m for pr, m in zip(prop_row, mask) if m]
            for param_row, prop_row in zip(self.params, self.results)
        ]
        table.set_data(data)
        return table.render()

    def _filter_properties(self, paths):
        def matches(column, path):
            for c, p in zip(column, path):
                if c != p:
                    return False
            return True

        def check(column):
            for path in paths:
                if matches(column, path):
                    return  True
            return False

        return [check(col) for col in self.result_columns]


class CaseStudy:
    def __init__(self, engine: CalculationEngine, scenario: Scenario,
                 out_stream):
        self.param_specs: list[ParameterSpec] = []
        self.results: Optional[CaseStudyResults] = None
        self.engine = engine
        self.outstream = out_stream
        self.lock = Lock()
        self.on_fail_continue: bool = True
        self.scenario = scenario
        self._interrupt = False

    def serialize(self, parameters):
        def serialize_param(p):
            p = dict(p)
            p["spec"] = p["spec"].serialize()
            p["min"] = f"{p['min']:.14g~}"
            p["max"] = f"{p['max']:.14g~}"
            p["units"] = list(p["units"])
            return p

        return {
            "scenario": self.scenario.serialize(),
            "parameters": list(map(serialize_param, parameters))
        }

    @classmethod
    def deserialize(cls, engine, outstream, data):
        def deserialize_param(p):
            return {
                "spec": ParameterSpec.deserialize(p["spec"]),
                "min": qty_cls(p["min"]),
                "max": qty_cls(p["min"]),
                "units": set(p["units"])
            }

        if data is None:
            return None
        qty_cls = get_unit_registry().Quantity
        scenario = Scenario.deserialize(data["scenario"])
        result = cls(engine, scenario, outstream)
        params = [deserialize_param(p) for p in data["parameters"]]
        result.param_specs = [p["spec"] for p in params]
        return result, params

    def collect(self, filename: str, paths: Sequence[Path]) -> str:
        return self.results.collect(filename, paths)

    def set_parameters(self, specs: Sequence[ParameterSpec]):
        self.param_specs = specs
        self.results = None

    def run(self):
        # make it possible to interrupt
        # fire events each time a result is obtained, and when it is ready
        def f():
            out = self.outstream
            data = [s.data for s in specs]
            for k, d in enumerate(product(*data), start=1):
                # set parameters
                print(f"Running case #{k}:", file=out)
                for n, p, d_i in zip(p_names, p_paths, d):
                    print(f"  {n} = {d_i:.6g~P}", file=out)
                    param.set(p, d_i)

                # run simulation
                try:
                    res = self.engine.calculate(param)
                except CalculationFailed as error:
                    pub.sendMessage(CALCULATION_FAILED, message=str(error))
                    if not self.on_fail_continue:
                        break
                else:
                    results.add_result(param, DataStructure(res))
                    pub.sendMessage(CASE_STUDY_PROGRESS, k=k)

                # catch if the case study was stopped.
                with self.lock:
                    interrupt = self._interrupt
                if interrupt:
                    break

            # send event of case study ended
            # if calculation was very fast, progress popup has not even
            # subscribed to events yet and will just hang, waiting for it.
            # waiting 200 ms prevents this with margin
            sleep(0.2)
            pub.sendMessage(CASE_STUDY_ENDED)

        self._interrupt = False
        param = deepcopy(self.scenario.parameters)
        specs = deepcopy(self.param_specs)

        p_paths = [p.path for p in specs]
        p_names = [p.name for p in specs]
        r_paths = self.scenario.results.all_paths
        self.results = results = CaseStudyResults(p_paths, r_paths)
        Thread(target=f, daemon=True).start()

    def interrupt(self):
        with self.lock:
            self._interrupt = True
